from __future__ import annotations

import collections
import logging
import queue
import random

from globus_compute_endpoint.logging_config import ComputeLogger

log: ComputeLogger = logging.getLogger(__name__)  # type: ignore
log.info("Interchange task dispatch started")


def naive_interchange_task_dispatch(
    interesting_managers: set[bytes],
    pending_task_queue: dict[str, queue.Queue[dict]],
    ready_manager_queue: dict[bytes, dict],
    scheduler_mode: str = "hard",
    cold_routing: bool = False,
) -> tuple[dict[bytes, list], int]:
    """
    This is an initial task dispatching algorithm for interchange.
    It returns a dictionary, whose key is manager, and the value is the list of tasks
    to be sent to manager, and the total number of dispatched tasks.
    """
    task_dispatch: dict[bytes, list] = {}
    dispatched_tasks = 0
    if scheduler_mode == "hard":
        dispatched_tasks += dispatch(
            task_dispatch,
            interesting_managers,
            pending_task_queue,
            ready_manager_queue,
            scheduler_mode="hard",
        )

    elif scheduler_mode == "soft":
        loops = ["warm"] if not cold_routing else ["warm", "cold"]
        for loop in loops:
            dispatched_tasks += dispatch(
                task_dispatch,
                interesting_managers,
                pending_task_queue,
                ready_manager_queue,
                scheduler_mode="soft",
                loop=loop,
            )
    return task_dispatch, dispatched_tasks


def dispatch(
    task_dispatch: dict[bytes, list],
    interesting_managers: set[bytes],
    pending_task_queue: dict[str, queue.Queue[dict]],
    ready_manager_queue: dict[bytes, dict],
    scheduler_mode: str = "hard",
    loop: str = "warm",
) -> int:
    """
    This is the core task dispatching algorithm for interchange.
    The algorithm depends on the scheduler mode and which loop.
    """
    dispatched_tasks = 0
    if interesting_managers:
        shuffled_managers = list(interesting_managers)
        random.shuffle(shuffled_managers)
        for manager in shuffled_managers:
            mdata = ready_manager_queue[manager]
            tasks_inflight = mdata["total_tasks"]
            real_capacity: int = min(
                mdata["free_capacity"]["total_workers"],
                mdata["max_worker_count"] - tasks_inflight,
            )
            if not (real_capacity > 0 and mdata["active"]):
                interesting_managers.remove(manager)
                continue

            if scheduler_mode == "hard":
                tasks, tids = get_tasks_hard(pending_task_queue, mdata, real_capacity)
            else:
                tasks, tids = get_tasks_soft(
                    pending_task_queue,
                    mdata,
                    real_capacity,
                    loop=loop,
                )
            if tasks:
                log.debug("Got %s tasks from queue", len(tasks))
                for task_type in tids:
                    # This line is a set update, not dict update
                    mdata["tasks"][task_type].update(tids[task_type])
                log.debug(f"The tasks on manager %s is {mdata['tasks']}", manager)
                mdata["total_tasks"] += len(tasks)
                if manager not in task_dispatch:
                    task_dispatch[manager] = []
                task_dispatch[manager] += tasks
                dispatched_tasks += len(tasks)
                log.debug("Assigned tasks %s to manager %s", tids, manager)
            if mdata["free_capacity"]["total_workers"] > 0:
                log.trace(
                    "Manager %s still has free_capacity %s",
                    manager,
                    mdata["free_capacity"]["total_workers"],
                )
            else:
                log.debug("Manager %s is now saturated", manager)
                interesting_managers.remove(manager)

    log.trace(
        "The task dispatch of %s loop is %s, in total %s tasks",
        loop,
        task_dispatch,
        dispatched_tasks,
    )
    return dispatched_tasks


def get_tasks_hard(
    pending_task_queue: dict[str, queue.Queue[dict]],
    manager_ads: dict,
    real_capacity: int,
) -> tuple[list[dict], dict[str, set[str]]]:
    tasks: list[dict] = []
    tids: dict[str, set[str]] = collections.defaultdict(set)
    task_type: str = manager_ads["worker_type"]
    if not task_type:
        log.warning(
            "Using hard scheduler mode but with manager worker type unset. "
            "Use soft scheduler mode. Set this in the config."
        )
        return tasks, tids
    task_q = pending_task_queue.get(task_type)
    if not task_q:
        log.trace("No task of type %s. Exiting task fetching.", task_type)
        return tasks, tids

    # dispatch tasks of available types on manager
    free_cap = manager_ads["free_capacity"]
    if task_type in free_cap["free"]:
        try:
            while real_capacity > 0 and free_cap["free"][task_type] > 0:
                x = task_q.get(block=False)
                log.debug(f"Get task {x}")
                tasks.append(x)
                tids[task_type].add(x["task_id"])
                free_cap["free"][task_type] -= 1
                free_cap["total_workers"] -= 1
                real_capacity -= 1
        except queue.Empty:
            pass

    # dispatch tasks to unused slots based on the manager type
    log.trace("Second round of task fetching in hard mode")
    try:
        while real_capacity > 0 and free_cap["free"]["unused"] > 0:
            x = task_q.get(block=False)
            log.debug(f"Get task {x}")
            tasks.append(x)
            tids[task_type].add(x["task_id"])
            free_cap["free"]["unused"] -= 1
            free_cap["total_workers"] -= 1
            real_capacity -= 1
    except queue.Empty:
        pass
    return tasks, tids


def get_tasks_soft(
    pending_task_queue: dict[str, queue.Queue[dict]],
    manager_ads: dict,
    real_capacity: int,
    loop: str = "warm",
) -> tuple[list[dict], dict[str, set[str]]]:
    tasks = []
    tids = collections.defaultdict(set)

    # Warm routing to dispatch tasks
    free_cap = manager_ads["free_capacity"]
    if loop == "warm":
        for task_type in free_cap["free"]:
            # Dispatch tasks that are of the available container types on the manager
            if task_type != "unused":
                task_q = pending_task_queue.get(task_type)
                if not task_q:
                    continue
                type_inflight = len(manager_ads["tasks"].get(task_type, set()))
                type_capacity = min(
                    free_cap["free"][task_type],
                    free_cap["total"][task_type] - type_inflight,
                )
                try:
                    while (
                        real_capacity > 0
                        and type_capacity > 0
                        and free_cap["free"][task_type] > 0
                    ):
                        x = task_q.get(block=False)
                        log.debug(f"Get task {x}")
                        tasks.append(x)
                        tids[task_type].add(x["task_id"])
                        free_cap["free"][task_type] -= 1
                        free_cap["total_workers"] -= 1
                        real_capacity -= 1
                        type_capacity -= 1
                except queue.Empty:
                    pass
            # Dispatch tasks to unused container slots on the manager
            else:
                task_q = pending_task_queue.get(task_type)  # "unused" queue
                if not task_q:
                    log.debug("Unexpectedly non-existent 'unused' queue")
                    continue
                task_types = list(pending_task_queue.keys())
                random.shuffle(task_types)
                for task_type in task_types:
                    try:
                        while (
                            real_capacity > 0
                            and free_cap["free"]["unused"] > 0
                            and free_cap["total_workers"] > 0
                        ):
                            x = task_q.get(block=False)
                            log.debug(f"Get task {x}")
                            tasks.append(x)
                            tids[task_type].add(x["task_id"])
                            free_cap["free"]["unused"] -= 1
                            free_cap["total_workers"] -= 1
                            real_capacity -= 1
                    except queue.Empty:
                        pass
        return tasks, tids

    # Cold routing round: allocate tasks of random types
    # to workers that are of different types on the manager
    # This will possibly cause container switching on the manager
    # This is needed to avoid workers being idle for too long
    # Potential issues may be that it could kill containers of short tasks frequently
    # Tune cold_routing_interval in the config to balance such a tradeoff
    log.debug("Cold function routing!")
    task_types = list(pending_task_queue.keys())
    random.shuffle(task_types)
    for task_type in task_types:
        task_q = pending_task_queue.get(task_type)
        if not task_q:
            continue
        try:
            while real_capacity > 0 and free_cap["total_workers"] > 0:
                x = task_q.get(block=False)
                tasks.append(x)
                tids[task_type].add(x["task_id"])
                free_cap["total_workers"] -= 1
                real_capacity -= 1
                log.debug(f"Get task {x}")
        except queue.Empty:
            pass
    return tasks, tids



from __future__ import annotations

import logging
import multiprocessing
import queue
import typing as t
import uuid
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor as NativeExecutor

import psutil
from globus_compute_common.messagepack.message_types import (
    EPStatusReport,
    TaskTransition,
)
from globus_compute_endpoint.engines.base import (
    GlobusComputeEngineBase,
    ReportingThread,
)

logger = logging.getLogger(__name__)


class ThreadPoolEngine(GlobusComputeEngineBase):
    def __init__(
        self,
        *args,
        label: str = "ThreadPoolEngine",
        heartbeat_period_s: float = 30.0,
        **kwargs,
    ):
        self.label = label
        self.executor = NativeExecutor(*args, **kwargs)
        self._status_report_thread = ReportingThread(
            target=self.report_status, args=[], reporting_period=heartbeat_period_s
        )
        super().__init__(*args, heartbeat_period_s=heartbeat_period_s, **kwargs)

    def start(
        self,
        *args,
        endpoint_id: t.Optional[uuid.UUID] = None,
        results_passthrough: t.Optional[queue.Queue] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        endpoint_id: Endpoint UUID
        results_passthrough: Queue to which packed results will be posted
        run_dir Not used
        Returns
        -------
        """
        assert endpoint_id, "ThreadPoolEngine requires kwarg:endpoint_id at start"
        self.endpoint_id = endpoint_id
        if results_passthrough:
            self.results_passthrough = results_passthrough
        assert self.results_passthrough

        # mypy think the thread can be none
        self._status_report_thread.start()

    def get_status_report(self) -> EPStatusReport:
        """
        endpoint_id: uuid.UUID
        ep_status_report: t.Dict[str, t.Any]
        task_statuses: t.Dict[str, t.List[TaskTransition]]
        Returns
        -------
        """
        executor_status: t.Dict[str, t.Any] = {
            "task_id": -2,
            "info": {
                "total_cores": multiprocessing.cpu_count(),
                "total_mem": round(psutil.virtual_memory().available / (2**30), 1),
                "total_core_hrs": 0,
                "total_workers": self.executor._max_workers,  # type: ignore
                "pending_tasks": 0,
                "outstanding_tasks": 0,
                "scaling_enabled": False,
                "max_blocks": 1,
                "min_blocks": 1,
                "max_workers_per_node": self.executor._max_workers,  # type: ignore
                "nodes_per_block": 1,
                "heartbeat_period": self._heartbeat_period_s,
            },
        }
        task_status_deltas: t.Dict[str, t.List[TaskTransition]] = {}

        return EPStatusReport(
            endpoint_id=self.endpoint_id,
            global_state=executor_status,
            task_statuses=task_status_deltas,
        )

    def _submit(
        self,
        func: t.Callable,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> Future:
        """We basically pass all params except the resource_specification
        over to executor.submit
        """
        logger.warning("Got task")
        return self.executor.submit(func, *args, **kwargs)

    def status_polling_interval(self) -> int:
        return 30

    def scale_out(self, blocks: int) -> list[str]:
        return []

    def scale_in(self, blocks: int) -> list[str]:
        return []

    def status(self) -> dict:
        return {}

    def shutdown(self, /, block=False, **kwargs) -> None:
        self._status_report_thread.stop()
        self.executor.shutdown(wait=block)



"""HighThroughputEngine builds on Parsl's HighThroughputExecutor for execution
of functions within containerized workers in a distributed setting.
"""
from __future__ import annotations

import concurrent.futures
import ipaddress
import logging
import multiprocessing
import os
import queue
import threading
import time
import typing as t
import uuid
from concurrent.futures import Future
from multiprocessing import Process

import dill
from globus_compute_common import messagepack
from globus_compute_endpoint.endpoint.messages_compat import convert_ep_status_report
from globus_compute_endpoint.engines.base import GlobusComputeEngineBase
from globus_compute_endpoint.engines.helper import execute_task
from globus_compute_endpoint.engines.high_throughput import interchange, zmq_pipes
from globus_compute_endpoint.engines.high_throughput.mac_safe_queue import mpQueue
from globus_compute_endpoint.engines.high_throughput.messages import (
    EPStatusReport,
    Heartbeat,
    HeartbeatReq,
    Task,
    TaskCancel,
)
from globus_compute_endpoint.strategies.simple import SimpleStrategy
from globus_compute_sdk.serialize import ComputeSerializer
from parsl.errors import ConfigurationError
from parsl.executors.errors import BadMessage, ScalingFailed
from parsl.providers import LocalProvider
from parsl.utils import RepresentationMixin

serializer = ComputeSerializer()

log = logging.getLogger(__name__)


BUFFER_THRESHOLD = 1024 * 1024
ITEM_THRESHOLD = 1024


class HighThroughputEngine(GlobusComputeEngineBase, RepresentationMixin):
    """Engine designed for cluster-scale

    The HighThroughputEngine system has the following components:
      1. The HighThroughputEngine instance which is run as a client
      2. The Interchange which is acts as a load-balancing proxy between workers and
         Parsl
      3. The multiprocessing based worker pool which coordinates task execution over
         several cores on a node.
      4. ZeroMQ pipes connect the HighThroughputEngine, Interchange and the
         process_worker_pool

    Here is a diagram

    .. code:: python


                        |  Data   |  Engine     |  Interchange  | External Process(es)
                        |  Flow   |             |               |
                   Task | Kernel  |             |               |
                 +----->|-------->|------------>|->outgoing_q---|-> process_worker_pool
                 |      |         |             | batching      |    |         |
           Parsl<---Fut-|         |             | load-balancing|  result   exception
                     ^  |         |             | watchdogs     |    |         |
                     |  |         |   Q_mngmnt  |               |    V         V
                     |  |         |    Thread<--|-incoming_q<---|--- +---------+
                     |  |         |      |      |               |
                     |  |         |      |      |               |
                     +----update_fut-----+


    Parameters
    ----------

    provider : :class:`~parsl.providers.base.ExecutionProvider`
       Provider to access computation resources. Can be one of
       :class:`~parsl.providers.aws.aws.EC2Provider`,
        :class:`~parsl.providers.cobalt.cobalt.Cobalt`,
        :class:`~parsl.providers.condor.condor.Condor`,
        :class:`~parsl.providers.googlecloud.googlecloud.GoogleCloud`,
        :class:`~parsl.providers.gridEngine.gridEngine.GridEngine`,
        :class:`~parsl.providers.jetstream.jetstream.Jetstream`,
        :class:`~parsl.providers.local.local.Local`,
        :class:`~parsl.providers.sge.sge.GridEngine`,
        :class:`~parsl.providers.slurm.slurm.Slurm`, or
        :class:`~parsl.providers.torque.torque.Torque`.

    label : str
        Label for this Engine instance.

    launch_cmd : str
        Command line string to launch the process_worker_pool from the provider. The
        command line string will be formatted with appropriate values for the following
        values: (
            debug,
            task_url,
            result_url,
            cores_per_worker,
            nodes_per_block,
            heartbeat_period,
            heartbeat_threshold,
            logdir,
        ).
        For example:
        launch_cmd="process_worker_pool.py {debug} -c {cores_per_worker} \
        --task_url={task_url} --result_url={result_url}"

    address : string
        An address of the host on which the engine runs, which is reachable from the
        network in which workers will be running. This can be either a hostname as
        returned by `hostname` or an IP address. Most login nodes on clusters have
        several network interfaces available, only some of which can be reached
        from the compute nodes. Some trial and error might be necessary to
        indentify what addresses are reachable from compute nodes.

    worker_ports : (int, int)
        Specify the ports to be used by workers to connect to Parsl. If this
        option is specified, worker_port_range will not be honored.

    worker_port_range : (int, int)
        Worker ports will be chosen between the two integers provided.

    interchange_port_range : (int, int)
        Port range used by Parsl to communicate with the Interchange.

    working_dir : str
        Working dir to be used by the engine.

    worker_debug : Bool
        Enables worker debug logging.

    cores_per_worker : float
        cores to be assigned to each worker. Oversubscription is possible
        by setting cores_per_worker < 1.0. Default=1

    mem_per_worker : float
        Memory to be assigned to each worker. Default=None(no limits)

    available_accelerators: int, list of str
        Either a list of accelerators device IDs
        or an integer defining the number of accelerators available.
        If an integer, sequential device IDs will be created starting at 0.
        The manager will ensure each worker is pinned to an accelerator and
        will set the maximum number of workers per node to be no more
        than the number of accelerators.

        Workers are pinned to specific accelerators using environment variables,
        such as by setting the ``CUDA_VISIBLE_DEVICES`` or
        ``SYCL_DEVICE_FILTER`` to the selected accelerator.
        Default: None

    max_workers_per_node : int
        Caps the number of workers launched by the manager. Default: infinity

    suppress_failure : Bool
        If set, the interchange will suppress failures rather than terminate early.
        Default: False

    heartbeat_threshold : int
        Seconds since the last message from the counterpart in the communication pair:
        (interchange, manager) after which the counterpart is assumed to be unavailable.
        Default:120s

    heartbeat_period : int
        Number of seconds after which a heartbeat message indicating liveness is sent to
        the endpoint
        counterpart (interchange, manager). Default:30s

    poll_period : int
        Timeout period to be used by the engine components in milliseconds.
        Increasing poll_periods trades performance for cpu efficiency. Default: 10ms

    container_image : str
        Path or identfier to the container image to be used by the workers

    scheduler_mode: str
        Scheduling mode to be used by the node manager. Options: 'hard', 'soft'
        'hard' -> managers cannot replace worker's container types
        'soft' -> managers can replace unused worker's containers based on demand

    worker_mode : str
        Select the mode of operation from no_container, singularity_reuse,
        singularity_single_use
        Default: singularity_reuse

    container_cmd_options: str
        Container command strings to be added to associated container command.
        For example, singularity exec {container_cmd_options}

    task_status_queue : queue.Queue
        Queue to pass updates to task statuses back to the forwarder.

    strategy: Stategy Object
        Specify the scaling strategy to use for this engine.

    launch_cmd: str
        Specify the launch command as using f-string format that will be used to specify
        command to launch managers. Default: None

    prefetch_capacity: int
        Number of tasks that can be fetched by managers in excess of available
        workers is a prefetching optimization. This option can cause poor
        load-balancing for long running functions.
        Default: 10

    provider: Provider object
        Provider determines how managers can be provisioned, say LocalProvider
        offers forked processes, and SlurmProvider interfaces to request
        resources from the Slurm batch scheduler.
        Default: LocalProvider
    """

    def __init__(
        self,
        label="HighThroughputEngine",
        # NEW
        strategy=SimpleStrategy(),
        max_workers_per_node=float("inf"),
        mem_per_worker=None,
        launch_cmd=None,
        available_accelerators=None,
        # Container specific
        worker_mode="no_container",
        scheduler_mode="hard",
        container_type=None,
        container_cmd_options="",
        cold_routing_interval=10.0,
        # Tuning info
        prefetch_capacity=10,
        provider=LocalProvider(),
        address="127.0.0.1",
        worker_ports=None,
        worker_port_range=(54000, 55000),
        interchange_port_range=(55000, 56000),
        storage_access=None,
        working_dir=None,
        worker_debug=False,
        cores_per_worker=1.0,
        heartbeat_threshold=120,
        heartbeat_period=30,
        poll_period=10,
        container_image=None,
        suppress_failure=True,
        run_dir=None,
        endpoint_id=None,
        passthrough=True,
        task_status_queue=None,
    ):
        log.debug("Initializing HighThroughputEngine")
        self.provider = provider
        self.label = label
        self.launch_cmd = launch_cmd
        self.worker_debug = worker_debug
        self.max_workers_per_node = max_workers_per_node

        # NEW
        self.strategy = strategy
        self.cores_per_worker = cores_per_worker
        self.mem_per_worker = mem_per_worker

        # Container specific
        self.scheduler_mode = scheduler_mode
        self.container_type = container_type
        self.container_cmd_options = container_cmd_options
        self.cold_routing_interval = cold_routing_interval

        # Tuning info
        self.prefetch_capacity = prefetch_capacity

        self.storage_access = storage_access if storage_access is not None else []
        if len(self.storage_access) > 1:
            raise ConfigurationError(
                "Multiple storage access schemes are not supported"
            )
        self.working_dir = working_dir
        self.blocks = []
        self.cores_per_worker = cores_per_worker
        self.endpoint_id = endpoint_id
        self._task_counter = 0

        try:
            ipaddress.ip_address(address=address)
        except Exception:
            log.critical(
                f"Invalid address supplied: {address}. "
                "Please use a valid IPv4 or IPv6 address"
            )
            raise
        self.address = address
        self.worker_ports = worker_ports
        self.worker_port_range = worker_port_range
        self.interchange_port_range = interchange_port_range
        self.heartbeat_threshold = heartbeat_threshold
        self.heartbeat_period = heartbeat_period
        self.poll_period = poll_period
        self.suppress_failure = suppress_failure
        self.run_dir = run_dir
        self.queue_proc: multiprocessing.Process | None = None
        self.passthrough = passthrough
        self.task_status_queue = task_status_queue
        self.tasks = {}

        self.outgoing_q: zmq_pipes.TasksOutgoing | None = None
        self.incoming_q: zmq_pipes.ResultsIncoming | None = None
        self.command_client: zmq_pipes.CommandClient | None = None
        self.results_passthrough: queue.Queue | None = None
        self._queue_management_thread: threading.Thread | None = None

        self.is_alive = False

        # Set the available accelerators
        if available_accelerators is None:
            self.available_accelerators = ()
        else:
            if isinstance(available_accelerators, int):
                self.available_accelerators = [
                    str(i) for i in range(available_accelerators)
                ]
            else:
                self.available_accelerators = list(available_accelerators)
            log.debug(
                "Workers will be assigned "
                f"to accelerators: {self.available_accelerators}"
            )

        # Globus Compute specific options
        self.container_image = container_image
        self.worker_mode = worker_mode

        if not launch_cmd:
            self.launch_cmd = (
                "process_worker_pool.py {debug} {max_workers} "
                "-c {cores_per_worker} "
                "--poll {poll_period} "
                "--task_url={task_url} "
                "--result_url={result_url} "
                "--logdir={logdir} "
                "--hb_period={heartbeat_period} "
                "--hb_threshold={heartbeat_threshold} "
                "--mode={worker_mode} "
                "--container_image={container_image} "
            )

    def start(
        self,
        *args,
        endpoint_id: uuid.UUID | None = None,
        run_dir: str | None = None,
        results_passthrough: queue.Queue[dict[str, bytes | str | None]] | None = None,
        **kwargs,
    ):
        """Create the Interchange process and connect to it."""
        assert run_dir, "HighThroughputEngine requires kwarg:run_dir at start"
        assert endpoint_id, "HighThroughputEngine requires kwarg:endpoint_id at start"
        self.run_dir = run_dir
        self.endpoint_id = endpoint_id

        self.outgoing_q = zmq_pipes.TasksOutgoing(
            "127.0.0.1", self.interchange_port_range
        )
        self.incoming_q = zmq_pipes.ResultsIncoming(
            "127.0.0.1", self.interchange_port_range
        )
        self.command_client = zmq_pipes.CommandClient(
            "127.0.0.1", self.interchange_port_range
        )

        self.is_alive = True

        if self.passthrough is True:
            if results_passthrough is None:
                raise Exception(
                    "Engines configured in passthrough mode, must be started with"
                    "a multiprocessing queue for results_passthrough"
                )
            self.results_passthrough = results_passthrough
            log.debug(f"Engine:{self.label} starting in results_passthrough mode")

        self._engine_bad_state = threading.Event()
        self._engine_exception: t.Optional[Exception] = None
        self._start_queue_management_thread()

        log.info("Attempting local interchange start")
        self._start_local_interchange_process()
        log.info(
            "Started local interchange with ports: %s. %s",
            self.worker_task_port,
            self.worker_result_port,
        )

        log.debug(f"Created management thread: {self._queue_management_thread}")

        if self.provider:
            pass
        else:
            self._scaling_enabled = False
            log.debug("Starting HighThroughputEngine with no provider")

        return self.outgoing_q.port, self.incoming_q.port, self.command_client.port

    def _start_local_interchange_process(self):
        """Starts the interchange process locally

        Starts the interchange process locally and uses an internal command queue to
        get the worker task and result ports that the interchange has bound to.
        """
        comm_q = mpQueue(maxsize=10)
        self.queue_proc = Process(
            target=interchange.starter,
            name="Engine-Interchange",
            args=(comm_q,),
            kwargs={
                "client_address": "127.0.0.1",  # engine and ix are on the same node
                "client_ports": (
                    self.outgoing_q.port,
                    self.incoming_q.port,
                    self.command_client.port,
                ),
                "provider": self.provider,
                "strategy": self.strategy,
                "poll_period": self.poll_period,
                "heartbeat_period": self.heartbeat_period,
                "heartbeat_threshold": self.heartbeat_threshold,
                "working_dir": self.working_dir,
                "worker_debug": self.worker_debug,
                "max_workers_per_node": self.max_workers_per_node,
                "mem_per_worker": self.mem_per_worker,
                "cores_per_worker": self.cores_per_worker,
                "available_accelerators": self.available_accelerators,
                "prefetch_capacity": self.prefetch_capacity,
                "scheduler_mode": self.scheduler_mode,
                "worker_mode": self.worker_mode,
                "container_type": self.container_type,
                "container_cmd_options": self.container_cmd_options,
                "cold_routing_interval": self.cold_routing_interval,
                "interchange_address": self.address,
                "worker_ports": self.worker_ports,
                "worker_port_range": self.worker_port_range,
                "logdir": os.path.join(self.run_dir, self.label),
                "suppress_failure": self.suppress_failure,
                "endpoint_id": self.endpoint_id,
            },
        )
        self.queue_proc.start()
        msg = None
        try:
            msg = comm_q.get(block=True, timeout=120)
        except queue.Empty:
            log.error("Interchange did not complete initialization.")

        if not msg:
            # poor-person's attempt to not interweave the traceback log lines
            # with the subprocess' likely traceback lines
            time.sleep(0.5)
            raise Exception("Interchange failed to start")

        comm_q.close()  # not strictly necessary, but be plain about intentions
        comm_q.join_thread()
        del comm_q

        self.worker_task_port, self.worker_result_port = msg

        self.worker_task_url = f"tcp://{self.address}:{self.worker_task_port}"
        self.worker_result_url = "tcp://{}:{}".format(
            self.address, self.worker_result_port
        )

    def get_status_report(self) -> EPStatusReport:
        """HTEX Interchange reports EPStatusReport periodically"""
        raise NotImplementedError

    def _queue_management_worker(self):
        """Listen to the queue for task status messages and handle them.

        Depending on the message, tasks will be updated with results, exceptions,
        or updates. It expects the following messages:

        .. code:: python

            {
               "task_id" : <task_id>
               "result"  : serialized result object, if task succeeded
               ... more tags could be added later
            }

            {
               "task_id" : <task_id>
               "exception" : serialized exception object, on failure
            }

        We do not support these yet, but they could be added easily.

        .. code:: python

            {
               "task_id" : <task_id>
               "cpu_stat" : <>
               "mem_stat" : <>
               "io_stat"  : <>
               "started"  : tstamp
            }

        The `None` message is a die request.
        """
        log.debug("queue management worker starting")

        while self.is_alive and not self._engine_bad_state.is_set():
            try:
                msgs = self.incoming_q.get(timeout=1)

            except queue.Empty:
                log.debug("queue empty")
                # Timed out.
                continue

            except OSError as e:
                log.exception(f"Caught broken queue with exception code {e.errno}: {e}")
                return

            except Exception as e:
                log.exception(f"Caught unknown exception: {e}")
                return

            else:
                if msgs is None:
                    log.debug("Got None, exiting")
                    return

                elif isinstance(msgs, EPStatusReport):
                    log.debug(f"Received {msgs!r}")
                    if self.passthrough:
                        external_ep_status = convert_ep_status_report(msgs)
                        self.results_passthrough.put(
                            {"message": messagepack.pack(external_ep_status)}
                        )

                else:
                    log.debug("Unpacking results")
                    for serialized_msg in msgs:
                        try:
                            msg = dill.loads(serialized_msg)
                            tid = msg["task_id"]
                        except dill.UnpicklingError:
                            raise BadMessage("Message received could not be unpickled")

                        except Exception:
                            raise BadMessage(
                                "Message received does not contain 'task_id' field"
                            )

                        if tid == -1 and "exception" in msg:
                            # TODO: This could be handled better we are
                            # essentially shutting down the client with little
                            # indication to the user.
                            log.warning(
                                "Engine shutting down due to fatal "
                                "exception from interchange"
                            )
                            self._engine_exception = serializer.deserialize(
                                msg["exception"]
                            )
                            log.exception(f"Exception: {self._engine_exception}")
                            # Set bad state to prevent new tasks from being submitted
                            self._engine_bad_state.set()
                            # We set all current tasks to this exception to make sure
                            # that this is raised in the main context.
                            # YADU: Report failure on all pending tasks
                            for task_id in self.tasks:
                                try:
                                    self.tasks[task_id].set_exception(
                                        self._engine_exception
                                    )
                                except concurrent.futures.InvalidStateError:
                                    # Task was already cancelled, the exception can be
                                    # ignored
                                    log.debug(
                                        f"Task:{task_id} result couldn't be set. "
                                        "Already in terminal state"
                                    )
                            break

                        if self.passthrough is True:
                            log.debug(f"Pushing results for task:{tid}")
                            # we are only interested in actual task ids here, not
                            # identifiers for other message types
                            sent_task_id = tid if isinstance(tid, str) else None
                            packed_result = self._convert_result_to_outgoing(msg)
                            task_result = {
                                "task_id": sent_task_id,
                                "message": packed_result,
                            }
                            self.results_passthrough.put(task_result)
                            self.tasks[sent_task_id].set_result(packed_result)
                            continue

                        try:
                            task_fut = self.tasks.pop(tid)
                        except KeyError:
                            # This is triggered when the result of a cancelled task is
                            # returned
                            # We should log, and proceed.
                            log.warning(
                                f"Task:{tid} not found in tasks table\n"
                                "Task likely was cancelled and removed."
                            )
                            continue

                        if "result" in msg:
                            result = serializer.deserialize(msg["result"])
                            try:
                                task_fut.set_result(result)
                            except concurrent.futures.InvalidStateError:
                                log.debug(
                                    f"Task:{tid} result couldn't be set. "
                                    "Already in terminal state"
                                )
                        elif "exception" in msg:
                            exception = serializer.deserialize(msg["exception"])
                            try:
                                task_fut.set_result(exception)
                            except concurrent.futures.InvalidStateError:
                                log.debug(
                                    f"Task:{tid} result couldn't be set. "
                                    "Already in terminal state"
                                )
                        else:
                            raise BadMessage(
                                "Message received is neither result or exception"
                            )

        log.info("queue management worker finished")

    def _convert_result_to_outgoing(self, msg: dict[str, t.Union[str, bytes]]) -> bytes:
        """The messagepacked result should be in the result dict in the data field"""
        # The data body already is packed on the worker
        assert isinstance(msg["data"], bytes)
        return msg["data"]

    # When the engine gets lost, the weakref callback will wake up
    # the queue management thread.
    def weakref_cb(self, q=None):
        """We do not use this yet."""
        q.put(None)

    def _start_queue_management_thread(self):
        """Method to start the management thread as a daemon.

        Checks if a thread already exists, then starts it.
        Could be used later as a restart if the management thread dies.
        """
        if self._queue_management_thread is None:
            log.debug("Starting queue management thread")
            self._queue_management_thread = threading.Thread(
                target=self._queue_management_worker, name="Queue-Management"
            )
            self._queue_management_thread.daemon = True
            self._queue_management_thread.start()
            log.debug("Started queue management thread")

        else:
            log.debug("Management thread already exists, returning")

    def hold_worker(self, worker_id):
        """Puts a worker on hold, preventing scheduling of additional tasks to it.

        This is called "hold" mostly because this only stops scheduling of tasks,
        and does not actually kill the worker.

        Parameters
        ----------

        worker_id : str
            Worker id to be put on hold
        """
        c = self.command_client.run(f"HOLD_WORKER;{worker_id}")
        log.debug(f"Sent hold request to worker: {worker_id}")
        return c

    def send_heartbeat(self):
        log.warning("Sending heartbeat to interchange")
        msg = Heartbeat(endpoint_id="")
        self.outgoing_q.put(msg.pack())

    def wait_for_endpoint(self):
        heartbeat = self.command_client.run(HeartbeatReq())
        log.debug("Attempting heartbeat to interchange")
        return heartbeat

    @property
    def outstanding(self):
        outstanding_c = self.command_client.run("OUTSTANDING_C")
        log.debug(f"Got outstanding count: {outstanding_c}")
        return outstanding_c

    @property
    def connected_workers(self):
        workers = self.command_client.run("MANAGERS")
        log.debug(f"Got managers: {workers}")
        return workers

    def _submit(self, func: t.Callable, *args: t.Any, **kwargs: t.Any) -> Future:
        raise RuntimeError("Invalid attempt to _submit()")

    def submit(self, task_id: str, packed_task: bytes) -> HTEXFuture:
        """Submits a messagepacked.Task for execution

        Parameters
        ----------
        Packed Task (messages.Task) - A packed Task object which contains task_id,
        container_id, and serialized fn, args, kwargs packages.

        Returns:
              Submit status
        """
        if self._engine_bad_state.is_set():
            # If the flag is set the exception body must exist
            raise self._engine_exception  # type: ignore

        self._task_counter += 1

        future = HTEXFuture(self, task_id)
        self.tasks[task_id] = future

        ser = serializer.serialize((execute_task, [task_id, packed_task], {}))
        payload = Task(task_id, "RAW", ser).pack()
        assert self.outgoing_q  # Placate mypy
        self.outgoing_q.put(payload)

        return future

    def _get_block_and_job_ids(self):
        # Not using self.blocks.keys() and self.blocks.values() simultaneously
        # The dictionary may be changed during invoking this function
        # As scale_in and scale_out are invoked in multiple threads
        block_ids = list(self.blocks.keys())
        job_ids = []  # types: List[Any]
        for bid in block_ids:
            job_ids.append(self.blocks[bid])
        return block_ids, job_ids

    @property
    def connection_info(self):
        """All connection info necessary for the endpoint to connect back

        Returns:
              Dict with connection info
        """
        return {
            "address": self.address,
            # A memorial to the ungodly amount of time and effort spent,
            # troubleshooting the order of these ports.
            "client_ports": "{},{},{}".format(
                self.outgoing_q.port, self.incoming_q.port, self.command_client.port
            ),
        }

    @property
    def scaling_enabled(self):
        return self._scaling_enabled

    def scale_out(self, blocks=1):
        """Scales out the number of blocks by "blocks"

        Raises:
             NotImplementedError
        """
        r = []
        for i in range(blocks):
            if self.provider:
                block = self.provider.submit(self.launch_cmd, 1, 1)
                log.debug(f"Launched block {i}:{block}")
                if not block:
                    raise (
                        ScalingFailed(
                            self.provider.label,
                            "Attempts to provision nodes via provider has failed",
                        )
                    )
                self.blocks.extend([block])
            else:
                log.error("No execution provider available")
                r = None
        return r

    def scale_in(self, blocks):
        """Scale in the number of active blocks by specified amount.

        The scale in method here is very rude. It doesn't give the workers
        the opportunity to finish current tasks or cleanup. This is tracked
        in issue #530

        Raises:
             NotImplementedError
        """
        to_kill = self.blocks[:blocks]
        if self.provider:
            r = self.provider.cancel(to_kill)
        return r

    def _get_job_ids(self):
        return list(self.block.values())

    def status(self):
        """Return status of all blocks."""

        status = []
        if self.provider:
            status = self.provider.status(self.blocks)

        return status

    def shutdown(self, /, hub=True, targets="all", block=False, **kwargs) -> None:
        """Shutdown the Engine, including all workers and controllers.

        This is not implemented.

        Kwargs:
            - hub (Bool): Whether the hub should be shutdown, Default:True,
            - targets (list of ints| 'all'): List of block id's to kill, Default:'all'
            - block (Bool): To block for confirmations or not

        Raises:
             NotImplementedError
        """

        log.info("Attempting HighThroughputEngine shutdown")
        if self.queue_proc:
            try:
                self.queue_proc.terminate()
            except AttributeError:
                log.info("Engine interchange terminate skipped due to wrong context")
            except Exception:
                log.exception("Terminating the interchange failed")
            if block:
                self.queue_proc.join(timeout=5)
                if self.queue_proc.exitcode is None:
                    self.queue_proc.kill()
            log.info("Finished HighThroughputEngine shutdown attempt")

    def cancel_task(self, task_id: uuid.UUID | str) -> bool:
        """
        Attempt to cancel the task `task_id` by requesting cancellation
        from the interchange.  Task cancellation is attempted only if the
        future is cancellable (i.e., not already in a terminal state).  This
        relies on the Engine not setting the task to a running state, and the
        task only tracking pending, and completed states.

        Parameters
        ----------
        task_id

        Returns
        -------
        Bool
        """

        log.debug("Send TaskCancel to interchange (%s)", task_id)
        log.debug("Sending cancel of task_id:{future.task_id} to interchange")
        # TODO: Doesn't yet return a bool ...
        assert self.command_client
        return self.command_client.run(TaskCancel(str(task_id)))


CANCELLED = "CANCELLED"
CANCELLED_AND_NOTIFIED = "CANCELLED_AND_NOTIFIED"
FINISHED = "FINISHED"


class HTEXFuture(concurrent.futures.Future):
    __slots__ = ("engine", "task_id")

    def __init__(self, engine: HighThroughputEngine, task_id: str):
        super().__init__()
        self.engine = engine
        self.task_id = task_id

    def cancel(self):
        raise NotImplementedError(
            f"{self.__class__} does not implement cancel() "
            "try using best_effort_cancel()"
        )

    def best_effort_cancel(self):
        """Attempt to cancel the function.

        If the function has finished running, the task cannot be cancelled
        and the method will return False.
        If the function is yet to start or is running, cancellation will be
        attempted without guarantees, and the method will return True.

        Please note that a return value of True does not guarantee that your
        function will not execute at all, but it does guarantee that the
        future will be in a cancelled state.

        Returns
        -------
        Bool
        """
        return self.engine.cancel_task(self.task_id)



from __future__ import annotations

import logging
import os
import queue
import random
import threading
import time
import typing as t

import pika
from globus_compute_endpoint.endpoint.utils import _redact_url_creds

if t.TYPE_CHECKING:
    from pika.channel import Channel
    from pika.frame import Method
    from pika.spec import Basic, BasicProperties

log = logging.getLogger(__name__)


class CommandQueueSubscriber(threading.Thread):
    def __init__(
        self,
        *,
        queue_info: dict,
        command_queue: queue.SimpleQueue[tuple[int, BasicProperties, bytes]],
        stop_event: threading.Event,
        poll_period_s: float = 0.5,
        connect_attempt_limit: int = 3,
        channel_close_window_s: int = 10,
        channel_close_window_limit: int = 3,
        thread_name: str | None = None,
    ):
        """
        Parameters
        ----------
        :param queue_info: the AMQP connection credentials, as received from upstream
        :param command_queue: Messages from upstream will be placed in this queue;
            consumers of this queue must call .to_ack() with the message id when
            finished processing
        :param stop_event: When set, the thread (CommandQueueSubscriber object) will
            shut down.  When the CQS' thread stops, this event will be summarily set.
            (Useful to know if the thread has shutdown prematurely.)
        :param poll_period_s: How often to perform housekeeping tasks (ACKing
            messages upstream, checking stop_event, etc.)
        :param connect_attempt_limit: Number of connection attempts to fail before
            giving up.  The connection counter will reset to 0 after the connection
            is sustained for 60s, so transient network errors should not build up
            to a future failure.)
        :param channel_close_window_s: Window of time to count channel close events
        :param channel_close_window_limit: Limit of channel close events (within
            ``channel_close_window_s``) before shutting down the thread.
        :param thread_name: Name the backing thread; per Python's implementation,
            this name has no semantics; default: implementation generated value.
        """
        super().__init__()

        self.queue_info = queue_info
        self._command_queue = command_queue
        self._stop_event = stop_event
        self._to_ack: queue.SimpleQueue[int] = queue.SimpleQueue()
        self._channel_closed = threading.Event()

        self._connection: pika.SelectConnection | None = None
        self._channel: Channel | None = None
        self._consumer_tag: str | None = None

        # how many times to attempt connection before giving up and shutting
        # down the thread
        self.connect_attempt_limit = connect_attempt_limit
        self._connection_tries = 0  # count of connection events; reset on success

        # invalid until set in start_consuming
        self._connected_at: int | None = None

        # list of times that channel was last closed
        self._channel_closes: list[float] = []

        # how long a time frame to keep previous channel close times
        self.channel_close_window_s = channel_close_window_s

        # how many times allowed to retry opening a channel in the above time
        # window before giving up and shutting down the thread
        self.channel_close_window_limit = channel_close_window_limit

        self._poll_period_s = poll_period_s
        if thread_name:
            self.name = thread_name

    def __repr__(self):
        return "{}<{}; pid={}>".format(
            self.__class__.__name__,
            "✓" if self._consumer_tag else "✗",
            os.getpid(),
        )

    def run(self):
        log.debug("%s AMQP thread begins", self)
        idle_for_s = 0.0
        while self._connection_tries < self.connect_attempt_limit and not (
            self._stop_event.is_set()
        ):
            if self._connection or self._connection_tries:
                idle_for_s = random.uniform(0.5, 10)
                msg = f"%s AMQP reconnecting in {idle_for_s:.1f}s."
                log.debug(msg, self)
                if self._connection_tries == self.connect_attempt_limit - 1:
                    log.warning(f"{msg}  (final attempt)", self)

            if self._stop_event.wait(idle_for_s):
                break

            self._connection_tries += 1
            try:
                log.debug(
                    "%s Opening connection to AMQP service.  Attempt: %s (of %s)",
                    self,
                    self._connection_tries,
                    self.connect_attempt_limit,
                )
                self._connection = self._connect()
                self._event_watcher()
                self._connection.ioloop.start()
            except Exception:
                log.exception("%s Unhandled exception: shutting down connection.", self)
        self._stop_event.set()
        log.debug("%s Shutdown complete", self)

    def _connect(self) -> pika.SelectConnection:
        pika_params = pika.URLParameters(self.queue_info["connection_url"])
        return pika.SelectConnection(
            pika_params,
            on_close_callback=self._on_connection_closed,
            on_open_error_callback=self._on_open_failed,
            on_open_callback=self._on_connection_open,
        )

    def _on_open_failed(self, mq_conn: pika.BaseConnection, exc: str | Exception):
        count = f"[attempt {self._connection_tries} (of {self.connect_attempt_limit})]"
        pid = f"(pid: {os.getpid()})"
        exc_text = f"Failed to open connection - ({exc.__class__.__name__}) {exc}"
        msg = f"{count} {pid} {exc_text}"
        log.debug("%s %s", self, msg)

        if not (self._connection_tries < self.connect_attempt_limit):
            self._stop_event.set()
            log.warning("%s %s", self, msg)
        mq_conn.ioloop.stop()

    def _on_connection_closed(self, mq_conn: pika.BaseConnection, exc: Exception):
        log.debug("%s Connection closed: %s", self, exc)
        self._consumer_tag = None
        mq_conn.ioloop.stop()

    def _on_connection_open(self, _mq_conn: pika.BaseConnection):
        log.debug("%s Connection established; creating channel", self)
        self._open_channel()

    def _open_channel(self):
        if self._connection and self._connection.is_open:
            self._connection.channel(on_open_callback=self._on_channel_open)

    def _on_channel_open(self, mq_chan: Channel):
        self._channel = mq_chan

        mq_chan.add_on_close_callback(self._on_channel_closed)
        mq_chan.add_on_cancel_callback(self._on_consumer_cancelled)

        log.debug(
            "%s Channel %s opened (%s)",
            self,
            mq_chan.channel_number,
            mq_chan.connection.params,
        )
        self._start_consuming()

    def _on_channel_closed(self, mq_chan: Channel, exc: Exception):
        self._consumer_tag = None
        now = time.monotonic()
        then = now - self.channel_close_window_s
        self._channel_closes = [cc for cc in self._channel_closes if cc > then]
        self._channel_closes.append(now)
        if len(self._channel_closes) < self.channel_close_window_limit:
            if self._stop_event.is_set():
                return
            msg = f"{self} Channel closed  [{mq_chan}\n  ({exc})]"
            log.debug(msg, exc_info=exc)
            log.warning(msg)
            mq_chan.connection.ioloop.call_later(1, self._open_channel)

        else:
            log.error(
                f"{self} Unable to sustain channel after {len(self._channel_closes)}"
                f" attempts in {self.channel_close_window_limit} seconds. ({exc})"
            )
            self._stop_event.set()

    def _on_consumer_cancelled(self, frame: Method[Basic.CancelOk]):
        log.debug("%s Consumer cancelled remotely, shutting down: %r", self, frame)
        if self._channel:
            self._channel.close()

    def _start_consuming(self):
        try:
            self._consumer_tag = self._channel.basic_consume(
                queue=self.queue_info["queue"],
                on_message_callback=self._on_message,
            )
            self._connected_at = time.time()
        except Exception as e:
            log.warning(
                f"{self} Unable to start consuming messages:"
                f" ({e.__class__.__name__}) {e}"
            )
            self._stop_ioloop()
        else:
            log.debug("%s Awaiting messages", self)

    def _on_message(
        self,
        mq_chan: Channel,
        basic_deliver: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
    ):
        try:
            d_tag = basic_deliver.delivery_tag
        except Exception as e:
            log.debug(
                "Invalid Basic.Deliver; unable to process message.  (%s) %s",
                e.__class__.__name__,
                e,
            )
            return

        try:
            log.debug(
                "%s Received message from %s: %s, %s",
                self,
                d_tag,
                properties.app_id,
                _redact_url_creds(body),
            )
            self._command_queue.put((d_tag, properties, body))
        except Exception:
            # No sense in waiting for the RMQ default 30m timeout; let it know
            # *now* that this message failed.
            log.exception("%s Command queue put failed", self)
            mq_chan.basic_nack(d_tag, requeue=True)

    def ack(self, msg_tag: int):
        self._to_ack.put(msg_tag)

    def _on_cancelok(self, _frame: Method[Basic.CancelOk]):
        self._close_channel()

    def _close_channel(self):
        log.debug("%s Closing the channel", self)
        self._channel.close()

    def _stop_ioloop(self):
        """
        Gracefully stop the ioloop.

        In an effort play nice with upstream, attempt to follow the AMQP protocol
        by closing the channel and connections gracefully.  This method will
        rearm itself while the connection is still open, continually working
        toward eventually and gracefully stopping the connection, before finally
        stopping the ioloop.
        """
        if self._connection:
            self._connection.ioloop.call_later(0.1, self._stop_ioloop)
            if self._connection.is_open:
                if self._channel:
                    if self._channel.is_open:
                        self._channel.close()
                    elif self._channel.is_closed:
                        self._channel = None
                else:
                    self._connection.close()
            elif self._connection.is_closed:
                self._connection.ioloop.stop()
                self._connection = None

    def _event_watcher(self):
        """Polls the stop_event periodically to trigger a shutdown"""
        if self._stop_event.is_set():
            log.debug("%s Shutting down per stop event", self)
            self._stop_ioloop()
            return

        if self._connection_tries and self._consumer_tag and self._connected_at:
            # we're connected ...
            if time.time() - self._connected_at > 60:
                # ... and connection stable for 60s; good to reset connection tries
                self._connection_tries = 0

        delivery_tags = []
        try:
            while True:
                delivery_tags.append(self._to_ack.get(block=False))
        except queue.Empty:
            pass
        if delivery_tags:
            delivery_tags.sort()  # nominally a no-op
            latest_msg_id = delivery_tags[-1]
            self._channel.basic_ack(latest_msg_id, multiple=True)
            log.debug("%s Acknowledged through message: %s", self, latest_msg_id)

        self._connection.ioloop.call_later(self._poll_period_s, self._event_watcher)



"""
This module provides helpers for constructing the internal dict objects which get passed
around and ultimately converted into funcx_common.messagepack.Result objects

IDEALLY this would be refactored to produce and return Result objects directly, which
could then be passed around internally instead of raw dicts
but we don't have time to do that (even though it would be better (a lot better))
"""

from __future__ import annotations

import functools
import logging
import os
import sys
import textwrap
import traceback
import types
import typing as t

import click
import globus_sdk
from globus_compute_endpoint.exceptions import CouldNotExecuteUserTaskError
from globus_compute_sdk.errors import MaxResultSizeExceeded

INTERNAL_ERROR_CLASSES: tuple[type[Exception], ...] = (
    CouldNotExecuteUserTaskError,
    MaxResultSizeExceeded,
)

log = logging.getLogger(__name__)


def _typed_excinfo(
    exc: Exception | None = None,
) -> tuple[type[Exception], Exception, types.TracebackType]:
    if exc:
        return type(exc), exc, exc.__traceback__  # type: ignore
    return t.cast(
        t.Tuple[t.Type[Exception], Exception, types.TracebackType],
        sys.exc_info(),
    )


def _inner_traceback(tb: types.TracebackType, levels: int = 2) -> types.TracebackType:
    while levels > 0:
        tb = tb.tb_next if tb.tb_next is not None else tb
        levels -= 1
    return tb


def get_error_string(*, exc: t.Any | None = None, tb_levels: int = 2) -> str:
    exc_info = _typed_excinfo(exc)
    exc_type, exc, tb = exc_info
    if isinstance(exc, INTERNAL_ERROR_CLASSES):
        return repr(exc)
    return "".join(
        traceback.format_exception(
            exc_type, exc, _inner_traceback(tb, levels=tb_levels)
        )
    )


def get_result_error_details(exc: BaseException | None = None) -> tuple[str, str]:
    _, error, _ = _typed_excinfo(exc)  # type: ignore
    # code, user_message
    if isinstance(error, INTERNAL_ERROR_CLASSES):
        return type(error).__name__, f"remote error: {error}"
    return (
        "RemoteExecutionError",
        "An error occurred during the execution of this task",
    )


def handle_auth_errors(f: t.Callable) -> t.Callable:
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except globus_sdk.AuthAPIError as e:
            msg = textwrap.dedent(
                f"""
                Compute CLI Error: An Auth API error occurred.
                HTTP status:       {e.http_status}
                code:              {e.code}
                message:           {e.text}
                """
            )
            log.warning(msg)
            click.echo(msg)
            sys.exit(os.EX_NOPERM)

    return wrapper



import logging
import os
import queue
import typing as t
import uuid
from concurrent.futures import Future

from globus_compute_common.messagepack.message_types import (
    EPStatusReport,
    TaskTransition,
)
from globus_compute_endpoint.engines.base import (
    GlobusComputeEngineBase,
    ReportingThread,
)
from globus_compute_endpoint.strategies import SimpleStrategy
from parsl.executors.high_throughput.executor import HighThroughputExecutor

logger = logging.getLogger(__name__)


class GlobusComputeEngine(GlobusComputeEngineBase):
    def __init__(
        self,
        *args,
        label: str = "GlobusComputeEngine",
        address: t.Optional[str] = None,
        heartbeat_period_s: float = 30.0,
        strategy: t.Optional[SimpleStrategy] = SimpleStrategy(),
        executor: t.Optional[HighThroughputExecutor] = None,
        **kwargs,
    ):
        self.address = address
        self.run_dir = os.getcwd()
        self.label = label
        self._status_report_thread = ReportingThread(
            target=self.report_status, args=[], reporting_period=heartbeat_period_s
        )
        super().__init__(*args, heartbeat_period_s=heartbeat_period_s, **kwargs)
        self.strategy = strategy
        self.max_workers_per_node = 1
        if executor is None:
            executor = HighThroughputExecutor(  # type: ignore
                *args, address=address, **kwargs
            )
        self.executor = executor

    def start(
        self,
        *args,
        endpoint_id: t.Optional[uuid.UUID] = None,
        run_dir: t.Optional[str] = None,
        results_passthrough: t.Optional[queue.Queue] = None,
        **kwargs,
    ):
        assert run_dir, "GCExecutor requires kwarg:run_dir at start"
        assert endpoint_id, "GCExecutor requires kwarg:endpoint_id at start"
        self.run_dir = os.path.join(os.getcwd(), run_dir)
        self.endpoint_id = endpoint_id
        script_dir = os.path.join(self.run_dir, "submit_scripts")
        self.executor.provider.script_dir = script_dir
        if (
            self.executor.provider.channel
            and not self.executor.provider.channel.script_dir
        ):
            self.executor.provider.channel.script_dir = script_dir

        os.makedirs(self.executor.provider.script_dir, exist_ok=True)
        if results_passthrough:
            # Only update the default queue in GCExecutorBase if
            # a queue is passed in
            self.results_passthrough = results_passthrough
        self.executor.start()
        if self.strategy:
            self.strategy.start(self)
        self._status_report_thread.start()

    def _submit(
        self,
        func: t.Callable,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> Future:
        return self.executor.submit(func, {}, *args, **kwargs)

    @property
    def provider(self):
        return self.executor.provider

    def get_outstanding_breakdown(self) -> t.List[t.Tuple[str, int, bool]]:
        """

        Returns
        -------
        List of tuples of the form (component, # of tasks on component, active?)
        """
        total_task_count = self.executor.outstanding
        manager_info: t.List[t.Dict[str, t.Any]] = self.executor.connected_managers()
        breakdown = [(m["manager"], m["tasks"], m["active"]) for m in manager_info]
        total_count_managers = sum([m["tasks"] for m in manager_info])
        task_count_interchange = total_task_count - total_count_managers
        breakdown = [("interchange", task_count_interchange, True)] + breakdown
        return breakdown

    def get_total_tasks_outstanding(self):
        """

        Returns
        -------
        Returns a dict of type {str_task_type: count_tasks}

        """
        outstanding = self.get_outstanding_breakdown()
        total = sum([component[1] for component in outstanding])
        return {"RAW": total}

    def provider_status(self):
        status = []
        if self.provider:
            # ex.locks is a dict of block_id:job_id mappings
            job_ids = self.executor.blocks.values()
            status = self.provider.status(job_ids=job_ids)
        return status

    def get_total_live_workers(self) -> int:
        manager_info: t.List[dict[str, t.Any]] = self.executor.connected_managers()
        worker_count = sum([mgr["worker_count"] for mgr in manager_info])
        return worker_count

    def scale_out(self, blocks: int):
        logger.info(f"Scaling out {blocks} blocks")
        return self.executor.scale_out(blocks=blocks)

    def scale_in(self, blocks: int):
        logger.info(f"Scaling in {blocks} blocks")
        to_kill = list(self.executor.blocks.values())[:blocks]
        return self.provider.cancel(to_kill)

    def get_status_report(self) -> EPStatusReport:
        """
        endpoint_id: uuid.UUID
        ep_status_report: t.Dict[str, t.Any]
        task_statuses: t.Dict[str, t.List[TaskTransition]]
        Returns
        -------
        """
        executor_status: t.Dict[str, t.Any] = {
            "task_id": -2,
            "info": {
                "total_cores": 0,
                "total_mem": 0,
                "new_core_hrs": 0,
                "total_core_hrs": 0,
                "managers": 0,
                "active_managers": 0,
                "total_workers": 0,
                "idle_workers": 0,
                "pending_tasks": 0,
                "outstanding_tasks": 0,
                "worker_mode": 0,
                "scheduler_mode": 0,
                "scaling_enabled": False,
                "mem_per_worker": 0,
                "cores_per_worker": 0,
                "prefetch_capacity": 0,
                "max_blocks": 1,
                "min_blocks": 1,
                "max_workers_per_node": 0,
                "nodes_per_block": 1,
                "heartbeat_period": self._heartbeat_period_s,
            },
        }
        task_status_deltas: t.Dict[str, t.List[TaskTransition]] = {}
        return EPStatusReport(
            endpoint_id=self.endpoint_id,
            global_state=executor_status,
            task_statuses=task_status_deltas,
        )

    def shutdown(self, /, **kwargs) -> None:
        self._status_report_thread.stop()
        if self.strategy:
            self.strategy.close()
        self.executor.shutdown()



from __future__ import annotations

import typing as t
from types import ModuleType

from globus_compute_endpoint import engines, strategies
from parsl import addresses as parsl_addresses
from parsl import channels as parsl_channels
from parsl import launchers as parsl_launchers
from parsl import providers as parsl_providers
from pydantic import BaseModel, validator


def _validate_import(field: str, package: ModuleType):
    def inner(cls, module: str):
        cls = getattr(package, module, None)
        if cls is None:
            raise ValueError(f"{module} could not be found")
        return cls

    return validator(field, allow_reuse=True)(inner)


def _validate_params(field: str):
    def inner(cls, model: t.Optional[BaseModel]):
        if not isinstance(model, BaseModel):
            return model

        fields = model.dict(exclude_none=True)
        cls = fields.pop("type")
        try:
            return cls(**fields)
        except Exception as err:
            raise ValueError(str(err)) from err

    return validator(field, allow_reuse=True)(inner)


class BaseConfigModel(BaseModel):
    class Config:
        extra = "allow"


class AddressModel(BaseConfigModel):
    type: str

    _validate_type = _validate_import("type", parsl_addresses)


class StrategyModel(BaseConfigModel):
    type: str

    _validate_type = _validate_import("type", strategies)


class LauncherModel(BaseConfigModel):
    type: str

    _validate_type = _validate_import("type", parsl_launchers)


class ChannelModel(BaseConfigModel):
    type: str

    _validate_type = _validate_import("type", parsl_channels)


class ProviderModel(BaseConfigModel):
    type: str
    channel: t.Optional[ChannelModel]
    launcher: t.Optional[LauncherModel]

    _validate_type = _validate_import("type", parsl_providers)
    _validate_channel = _validate_params("channel")
    _validate_launcher = _validate_params("launcher")


class EngineModel(BaseConfigModel):
    type: str = "HighThroughputEngine"
    provider: t.Optional[ProviderModel]
    strategy: t.Optional[StrategyModel]
    address: t.Optional[t.Union[str, AddressModel]]

    _validate_type = _validate_import("type", engines)
    _validate_provider = _validate_params("provider")
    _validate_strategy = _validate_params("strategy")
    _validate_address = _validate_params("address")

    class Config:
        validate_all = True


class ConfigModel(BaseConfigModel):
    engine: EngineModel
    display_name: t.Optional[str]
    environment: t.Optional[str]
    funcx_service_address: t.Optional[str]
    multi_tenant: t.Optional[bool]
    allowed_functions: t.Optional[t.List[str]]
    heartbeat_period: t.Optional[int]
    heartbeat_threshold: t.Optional[int]
    idle_heartbeats_soft: t.Optional[int]
    idle_heartbeats_hard: t.Optional[int]
    detach_endpoint: t.Optional[bool]
    endpoint_setup: t.Optional[str]
    endpoint_teardown: t.Optional[str]
    log_dir: t.Optional[str]
    stdout: t.Optional[str]
    stderr: t.Optional[str]

    _validate_engine = _validate_params("engine")

    def dict(self, *args, **kwargs):
        # Slight modification is needed here since we still
        # store the engine/executor in a list named executors
        ret = super().dict(*args, **kwargs)
        executor = ret.pop("engine")
        ret["executors"] = [executor]
        return ret



from __future__ import annotations

import glob
import os
import shlex
import shutil
import socket
import ssl
import subprocess
import sys
import textwrap
from urllib.parse import urlparse

import click
from globus_compute_sdk.sdk._environments import (
    get_amqp_service_host,
    get_web_service_url,
)
from globus_compute_sdk.sdk.web_client import WebClient


def cat(path: str, wildcard: bool = False, max_bytes: int | None = None):
    def kernel() -> None:
        full_path = os.path.expanduser(path)

        if wildcard:
            files = glob.glob(full_path)
        else:
            files = [full_path]

        for filename in files:
            click.echo("cat " + filename)
            if os.path.exists(filename):
                with open(filename, "rb") as f:
                    if max_bytes:
                        file_size = os.fstat(f.fileno()).st_size
                        f.seek(max(file_size - max_bytes, 0))
                    content = f.read().decode("utf-8")
                click.echo(textwrap.indent(content, "  "))
            else:
                click.secho(f"No file named {filename}\n", fg="red", bold=True)

    kernel.display_name = f"func:cat({path})"  # type: ignore
    return kernel


def test_conn(host: str, port: int, timeout: int = 5):
    def kernel():
        try:
            socket.create_connection((host, port), timeout=timeout)
            click.echo(f"Connected successfully to {host} over port {port}!\n")
        except OSError as e:
            click.secho(
                f"Connection failed to {host} over port {port}: {e}\n",
                fg="red",
                bold=True,
            )

    kernel.display_name = f"func:test_conn({host}, {port})"  # type: ignore
    return kernel


def test_ssl_conn(host: str, port: int, timeout: int = 5):
    def kernel() -> None:
        context = ssl.create_default_context()
        try:
            with socket.create_connection((host, port), timeout=timeout) as sock:
                with context.wrap_socket(sock, server_hostname=host) as ssock:
                    conn_data = ssock.cipher()
                    cipher, version, _ = conn_data if conn_data else (None, None, None)
            click.echo(
                f"Successfully established SSL connection with {host}:{port}!\n"
                f"Version: {version}\n"
                f"Cipher:  {cipher}\n"
            )
        except OSError as e:
            click.secho(
                f"Failed to establish SSL connection with {host}:{port} - {e}\n",
                fg="red",
                bold=True,
            )

    kernel.display_name = f"func:test_ssl_conn({host}, {port})"  # type: ignore
    return kernel


def get_service_versions(base_url: str):
    def kernel():
        wc = WebClient(base_url=base_url)
        res = wc.get_version(service="all")
        click.echo(f"{res}\n")

    kernel.display_name = f"func:get_service_versions({base_url})"  # type: ignore
    return kernel


def get_openssl_version():
    click.echo(f"{ssl.OPENSSL_VERSION}\n")


def get_python_version():
    click.echo(f"Python version {sys.version}\n")


def which_python():
    click.echo(f"{sys.executable}\n")


def _run_command(cmd: str):
    cmd_list = shlex.split(cmd)
    arg0 = cmd_list[0]

    if not shutil.which(arg0):
        click.secho(f"{arg0} was not found in the PATH\n", fg="red", bold=True)
        return

    try:
        res = subprocess.run(cmd_list, timeout=30, capture_output=True)
        if res.stdout:
            click.echo(res.stdout)
        if res.stderr:
            click.secho(res.stderr.decode("utf-8"), fg="red", bold=True)
    except subprocess.TimeoutExpired:
        click.secho("Command timed out\n", fg="red", bold=True)
    except Exception as e:
        click.secho(f"Command failed: {e}\n", fg="red", bold=True)


def run_self_diagnostic(log_bytes: int | None = None):
    web_svc_url = get_web_service_url()
    web_svc_host = urlparse(web_svc_url).netloc
    amqp_svc_host = get_amqp_service_host()

    commands = [
        "uname -a",
        cat("/etc/os-release"),
        "whoami",
        which_python,
        get_python_version,
        "pip freeze",
        get_openssl_version,
        test_conn(web_svc_host, 443),
        test_conn(amqp_svc_host, 5671),
        test_ssl_conn(web_svc_host, 443),
        test_ssl_conn(amqp_svc_host, 5671),
        get_service_versions(web_svc_url),
        "ip addr",
        "ifconfig",
        "ip route",
        "netstat -r",
        "globus-compute-endpoint whoami",
        "globus-compute-endpoint list",
        cat("~/.globus_compute/*/config.*", wildcard=True),
        cat("~/.globus_compute/*/endpoint.log", wildcard=True, max_bytes=log_bytes),
    ]

    for cmd in commands:
        display_name = (
            str(cmd)
            if not callable(cmd)
            else getattr(cmd, "display_name", f"func:{cmd.__name__}()")
        )
        click.secho(f"== Diagnostic: {display_name} ==", fg="yellow", bold=True)

        if callable(cmd):
            cmd()
            continue

        _run_command(cmd)



import os as _os
import pwd as _pwd
import re as _re
import typing as t

try:
    import pyprctl as _pyprctl

    has_pyprctl = True
except AttributeError:
    has_pyprctl = False


_url_user_pass_pattern = r"://([^:]+):[^@]+@"
_url_user_pass_re = _re.compile(_url_user_pass_pattern)
_urlb_user_pass_re = _re.compile(_url_user_pass_pattern.encode())

_T = t.TypeVar("_T", str, bytes)

if has_pyprctl:
    # List of Linux capabilities that would suggest the running process is
    # privileged, even if os.getuid() != 0.
    _MULTI_USER_CAPS = {
        _pyprctl.Cap.AUDIT_CONTROL,
        _pyprctl.Cap.AUDIT_READ,
        _pyprctl.Cap.BPF,
        _pyprctl.Cap.CHECKPOINT_RESTORE,
        _pyprctl.Cap.CHOWN,
        _pyprctl.Cap.DAC_OVERRIDE,
        _pyprctl.Cap.DAC_READ_SEARCH,
        _pyprctl.Cap.FOWNER,
        _pyprctl.Cap.FSETID,
        _pyprctl.Cap.IPC_OWNER,
        _pyprctl.Cap.KILL,
        _pyprctl.Cap.LINUX_IMMUTABLE,
        _pyprctl.Cap.MAC_ADMIN,
        _pyprctl.Cap.MAC_OVERRIDE,
        _pyprctl.Cap.MKNOD,
        _pyprctl.Cap.NET_ADMIN,
        _pyprctl.Cap.NET_BIND_SERVICE,
        _pyprctl.Cap.NET_RAW,
        _pyprctl.Cap.PERFMON,
        _pyprctl.Cap.SETGID,
        _pyprctl.Cap.SETFCAP,
        _pyprctl.Cap.SETPCAP,
        _pyprctl.Cap.SETUID,
        _pyprctl.Cap.SYS_ADMIN,
        _pyprctl.Cap.SYS_BOOT,
        _pyprctl.Cap.SYS_CHROOT,
        _pyprctl.Cap.SYS_MODULE,
        _pyprctl.Cap.SYS_NICE,
        _pyprctl.Cap.SYS_PACCT,
        _pyprctl.Cap.SYS_PTRACE,
        _pyprctl.Cap.SYS_RAWIO,
        _pyprctl.Cap.SYS_RESOURCE,
        _pyprctl.Cap.SYS_TIME,
        _pyprctl.Cap.SYS_TTY_CONFIG,
        _pyprctl.Cap.SYSLOG,
        _pyprctl.Cap.WAKE_ALARM,
    }


def _redact_url_creds(raw: _T, redact_user=True, repl="***", count=0) -> _T:
    """
    Redact URL credentials found in `raw`, by replacing the password and
    (optionally) username with `repl`.

    (A wrapper over `re.sub()`)

    :param raw: The raw string to be redacted.
    :param redact_user: If false, do not redact the username
    :param count: If 0, replace all found occurrences; otherwise, replace only count
    :return: A string (or bytes) with URL credentials redacted
    """
    if redact_user:
        repl = rf"://{repl}:{repl}@"
    else:
        repl = rf"://\1:{repl}@"
    if isinstance(raw, str):
        return _url_user_pass_re.sub(repl=repl, string=raw, count=count)
    return _urlb_user_pass_re.sub(repl=repl.encode(), string=raw, count=count)


def is_privileged(posix_user=None):
    if not posix_user:
        posix_user = _pwd.getpwuid(_os.getuid())

    proc_caps = _pyprctl.CapState.get_current()
    has_privileges = posix_user.pw_uid == 0
    has_privileges |= posix_user.pw_name == "root"
    has_privileges |= any(c in proc_caps.effective for c in _MULTI_USER_CAPS)
    return has_privileges



import warnings

from globus_compute_endpoint.engines import HighThroughputEngine

warnings.warn(
    f"{__name__} is deprecated.  Please import from globus_compute_sdk.engines instead"
)


def HighThroughputExecutor(*args, **kwargs) -> HighThroughputEngine:
    warnings.warn(
        "HighThroughputExecutor is deprecated.  Please use HighThroughputEngine instead"
    )
    return HighThroughputEngine(*args, **kwargs)



from __future__ import annotations

import logging
import multiprocessing
import signal

# multiprocessing.Event is a method, not a class
# to annotate, we need the "real" class
# see: https://github.com/python/typeshed/issues/4266
from multiprocessing.synchronize import Event as EventType

import pika
import pika.channel
import pika.exceptions
import pika.frame

from .base import SubscriberProcessStatus

logger = logging.getLogger(__name__)


class TaskQueueSubscriber(multiprocessing.Process):
    """The TaskQueueSubscriber is a direct rabbitMQ pipe subscriber that uses
    the SelectConnection adaptor to enable performance consumption of messages
    from the service

    """

    EXCHANGE_NAME = "tasks"
    EXCHANGE_TYPE = "direct"

    def __init__(
        self,
        *,
        endpoint_id: str,
        queue_info: dict,
        external_queue: multiprocessing.Queue,
        quiesce_event: EventType,
    ):
        """

        Parameters
        ----------
        queue_info: Queue information
             Dictionary that includes the key "connection_url", as well as exchange
             and queue declaration information specified by the server.

        external_queue: multiprocessing.Queue
             Each incoming message will be pushed to the queue.
             Please note that upon pushing a message into this queue, it will be
             marked as delivered.

        quiesce_event: multiprocessing.Event
             This event is used to communicate a failure on the subscriber

        endpoint_id: endpoint uuid string
        """

        super().__init__()
        self.status = SubscriberProcessStatus.parent

        self.endpoint_id = endpoint_id
        self.external_queue = external_queue
        self.queue_info = queue_info
        self.quiesce_event = quiesce_event

        self._channel_closed = multiprocessing.Event()
        self._cleanup_complete = multiprocessing.Event()

        self._connection: pika.SelectConnection | None = None
        self._channel: pika.channel.Channel | None = None
        self._consumer_tag: str | None = None

        self._watcher_poll_period = 0.1  # seconds
        logger.debug("Init done")

    def _connect(self) -> pika.SelectConnection:
        pika_params = pika.URLParameters(self.queue_info["connection_url"])
        return pika.SelectConnection(
            pika_params,
            on_open_callback=self._on_connection_open,
            on_close_callback=self._on_connection_closed,
            on_open_error_callback=self._on_open_failed,
        )

    def _on_connection_open(self, _unused_connection):
        logger.info("Connection opened")
        self.open_channel()

    def _on_open_failed(self, *args):
        logger.warning(f"Connection failed to open : {args}")

    def _on_connection_closed(self, connection, exception):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param exception: Exception object
        """
        logger.warning(f"Connection closing: {exception}")
        self._channel = None
        # Setting the quiesce_event will trigger shutdown via the event_watcher
        self.quiesce_event.set()

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if self.status is SubscriberProcessStatus.running:
            # Create a new connection
            self._connection = self._connect()

            # There is now a new connection, needs a new ioloop to run
            self._connection.ioloop.start()

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        logger.info("Creating a new channel")
        self._connection.channel(on_open_callback=self._on_channel_open)

    def _on_channel_open(self, channel: pika.channel.Channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the EXCHANGE_NAME to use.

        :param pika.channel.Channel channel: The channel object

        """
        logger.info(f"Channel opened: {channel}")
        self._channel = channel

        logger.debug("Adding channel close callback")
        channel.add_on_close_callback(self._on_channel_closed)

        logger.debug("Adding consumer cancellation callback")
        channel.add_on_cancel_callback(self.on_consumer_cancelled)

        logger.info(f"Ensuring exchange exists: {self.queue_info['exchange']}")
        channel.exchange_declare(
            passive=True,  # *we* don't create the exchange
            callback=self._on_exchange_declareok,
            exchange=self.queue_info["exchange"],
        )

    def _on_channel_closed(self, channel, exception):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an EXCHANGE_NAME or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        This is also invoked at the end of a successful client-initiated close, as when
        `connection.close()` is called and closes any open channels.

        :param pika.channel.Channel: The closed channel
        :param Exception: exception from pika
        """
        logger.warning(f"Channel:{channel} was closed: ({exception}")
        if isinstance(exception, pika.exceptions.ChannelClosedByBroker):
            logger.warning("Channel closed by RabbitMQ broker")
            if "exclusive use" in exception.reply_text:
                logger.exception(
                    "Channel closed by RabbitMQ broker due to exclusive "
                    "ownership by an active endpoint"
                )
                logger.warning("Channel will close without connection retry")
                self.status = SubscriberProcessStatus.closing
                self.quiesce_event.set()
        elif isinstance(exception, pika.exceptions.ChannelClosedByClient):
            logger.debug("Detected channel closed by client")
        else:
            logger.exception("Channel closed by unhandled exception.")
        logger.debug("marking channel as closed")
        self._channel_closed.set()

    def _on_exchange_declareok(self, _frame: pika.frame.Method):
        """
        Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC command.
        """
        logger.info(
            "Exchange declared successfully.  Ensuring queue exists:"
            f" {self.queue_info['queue']}"
        )
        assert self._channel is not None
        self._channel.queue_declare(
            passive=True,  # *we* don't create the queue, just consume it
            callback=self._on_queue_declareok,
            queue=self.queue_info["queue"],
        )

    def _on_queue_declareok(self, _frame: pika.frame.Method):
        self.start_consuming()

    def start_consuming(self):
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.

        """
        logger.info("Issuing consumer related RPC commands")
        self._consumer_tag = self._channel.basic_consume(
            queue=self.queue_info["queue"],
            on_message_callback=self.on_message,
            exclusive=True,
        )

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        logger.info("Consumer was cancelled remotely, shutting down: %r", method_frame)
        if self._channel:
            self._channel.close()

    def on_message(
        self,
        channel: pika.channel.Channel,
        basic_deliver: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the EXCHANGE_NAME, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.
        """
        logger.debug(
            "Received message from %s: %s, %s",
            basic_deliver.delivery_tag,
            properties.app_id,
            body,
        )

        # Not sure if we need to do this in a locked context,
        # rabbit's ACK system should make sure you don't lose tasks.
        try:
            # TODO move ack'ing the message to final processing instead
            headers = properties.headers if properties.headers else {}
            self.external_queue.put([headers, body])
        except Exception:
            # No sense in waiting for the RMQ default 30m timeout; let it know
            # *now* that this message failed.
            logger.exception("External queue put failed")
            channel.basic_nack(basic_deliver.delivery_tag, requeue=True)
        else:
            channel.basic_ack(basic_deliver.delivery_tag)
            logger.debug("Acknowledged message: %s", basic_deliver.delivery_tag)

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.

        """
        if self._channel:
            logger.info("Sending a Basic.Cancel RPC command to RabbitMQ")
            self._channel.basic_cancel(
                consumer_tag=self._consumer_tag, callback=self.on_cancelok
            )

    def on_cancelok(self, unused_frame):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        :param pika.frame.Method unused_frame: The Basic.CancelOk frame

        """
        logger.info("RabbitMQ acknowledged the cancellation of the consumer")
        self.close_channel()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        logger.info("Closing the channel")
        self._channel.close()

    def _shutdown(self):
        logger.debug("set status to 'closing'")
        self.status = SubscriberProcessStatus.closing

        if self._connection and self._connection.is_open:
            logger.debug("closing connection")
            try:
                self._connection.close()
            except Exception as exc:
                logger.warning(f"Unexpected error while closing connection: {exc}")

        logger.debug("stopping ioloop")
        self._connection.ioloop.stop()
        logger.debug("waiting until channel is closed (timeout=1 second)")
        if not self._channel_closed.wait(1.0):
            logger.warning("reached timeout while waiting for channel closed")
        logger.debug("closing connection to mp queue")
        self.external_queue.close()
        logger.debug("joining mp queue background thread")
        self.external_queue.join_thread()
        logger.info("shutdown done, setting cleanup event")
        self._cleanup_complete.set()

    def event_watcher(self):
        """Polls the quiesce_event periodically to trigger a shutdown"""
        if self.quiesce_event.is_set():
            logger.info("Shutting down task queue reader due to quiesce event.")
            try:
                self._shutdown()
            except Exception:
                logger.exception("error while shutting down")
                raise
            logger.info("Shutdown complete")
        else:
            self._connection.ioloop.call_later(
                self._watcher_poll_period, self.event_watcher
            )

    def handle_sigterm(self, sig_num, curr_stack_frame):
        logger.warning("Received SIGTERM, setting stop event")
        self.quiesce_event.set()

    def run(self):
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to
        operate.

        Note: Only one of these options should be used.
        """
        signal.signal(signal.SIGTERM, self.handle_sigterm)
        self.status = SubscriberProcessStatus.running
        try:
            self._connection = self._connect()
            self.event_watcher()
            self._connection.ioloop.start()
        except Exception:
            logger.exception("Failed to start subscriber")
            self.quiesce_event.set()

    def stop(self) -> None:
        """stop() is called by the parent to shutdown the subscriber"""
        logger.info("Stopping")
        self.quiesce_event.set()
        logger.info("Waiting for cleanup_complete")
        if not self._cleanup_complete.wait(2 * self._watcher_poll_period):
            logger.warning("Reached timeout while waiting for cleanup complete")
        # join shouldn't block if the above did not raise a timeout
        self.join()
        self.close()
        logger.info("Cleanup done")



"""
This module contains logging configuration for the globus-compute-endpoint application.
"""
from __future__ import annotations

import copy
import logging
import logging.config
import logging.handlers
import os
import pathlib
import re
import sys
import uuid
from datetime import datetime

log = logging.getLogger(__name__)

DEFAULT_FORMAT = (
    "%(asctime)s %(levelname)s %(processName)s-%(process)d "
    "%(threadName)s-%(thread)d %(name)s:%(lineno)d %(funcName)s "
    "%(message)s"
)

_ital = "\033[3m"
_redb = "\033[41m"
_teal = "\033[32m"
_yel = "\033[33m"
_byel = "\033[93m"
_yelb = "\033[43m"
_purp = "\033[35m"
_cyan = "\033[36m"
_gray = "\033[37m"
_grayonb = "\033[37;40m"
_r = "\033[m"
_C_BASE = (
    f"{_yel}%(asctime)s{_r} {_ital}%(levelname)s{_r}"
    " %(processName)s-%(process)d %(threadName)s-%(thread)d"
    f" %(name)s:{_cyan}%(lineno)d{_r} {_purp}%(funcName)s{_r}"
)
COLOR_ERROR = _redb
COLOR_WARNING = _yelb
COLOR_INFO = _gray
COLOR_DEBUG = _grayonb
C_ERROR_FMT = _C_BASE + f" {COLOR_ERROR}%(message)s{_r}"
C_WARNING_FMT = _C_BASE + f" {COLOR_WARNING}%(message)s{_r}"
C_INFO_FMT = _C_BASE + f" {COLOR_INFO}%(message)s{_r}"
C_DEBUG_FMT = _C_BASE + f" {COLOR_DEBUG}%(message)s{_r}"


class DatetimeFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        ct = datetime.fromtimestamp(record.created)
        if not datefmt:
            datefmt = self.default_time_format
        return ct.strftime(datefmt)


class ComputeConsoleFormatter(logging.Formatter):
    """
    For internal use only.
    This formatter handles output to standard streams in the following way:

    if 'debug' is False (default):
        info messages and below are treated as "user output" and are minimally decorated
        warning messages and up are given a full format, so are debug messages

    if 'debug' is True:
        all messages are given the full format
    """

    _u = "[0-9A-Fa-f]"  # convenience
    _uuid_re = f"{_u}{{8}}-{_u}{{4}}-{_u}{{4}}-{_u}{{4}}-{_u}{{12}}"
    # match uuids for colorization that have not otherwise already been colorized
    uuid_re = re.compile(rf"(?<!\dm)({_uuid_re})")

    def __init__(
        self, debug: bool = False, no_color: bool = False, fmt: str = "", **kwargs
    ) -> None:
        super().__init__()

        kwargs.setdefault("datefmt", "%Y-%m-%d %H:%M:%S,%f")

        self.use_color = debug and not no_color and sys.stderr.isatty()

        if fmt:
            d_fmt, i_fmt, w_fmt, e_fmt = fmt, fmt, fmt, fmt
        else:
            d_fmt, i_fmt, w_fmt, e_fmt = (
                C_DEBUG_FMT,
                C_INFO_FMT,
                C_WARNING_FMT,
                C_ERROR_FMT,
            )

        if not self.use_color:
            ansi_re = re.compile("\033.*?m")
            d_fmt = ansi_re.sub("", d_fmt)
            i_fmt = ansi_re.sub("", i_fmt)
            w_fmt = ansi_re.sub("", w_fmt)
            e_fmt = ansi_re.sub("", e_fmt)

        if debug:
            self._error_formatter = DatetimeFormatter(fmt=e_fmt, **kwargs)
            self._warning_formatter = DatetimeFormatter(fmt=w_fmt, **kwargs)
            self._debug_formatter = DatetimeFormatter(fmt=d_fmt, **kwargs)
            self._info_formatter = DatetimeFormatter(fmt=i_fmt, **kwargs)
        else:
            self._info_formatter = DatetimeFormatter(fmt="> %(message)s")
            self._warning_formatter = DatetimeFormatter(fmt=w_fmt, **kwargs)
            self._error_formatter = self._warning_formatter
            self._debug_formatter = self._warning_formatter

    def format(self, record: logging.LogRecord):
        if self.use_color:
            # Highlight all UUIDs
            if record.levelno > logging.WARNING:
                end_coloring = COLOR_ERROR
            elif record.levelno > logging.INFO:
                end_coloring = COLOR_WARNING
            elif record.levelno > logging.DEBUG:
                end_coloring = COLOR_INFO
            else:
                end_coloring = COLOR_DEBUG

            repl = f"{_byel}\\1{_r}{end_coloring}"
            try:
                record.msg = self.uuid_re.sub(repl, record.msg)
                if isinstance(record.args, dict):
                    record.args = copy.deepcopy(record.args)
                    for k, v in record.args.items():
                        if isinstance(v, (str, uuid.UUID)):
                            record.args[k] = self.uuid_re.sub(repl, str(v))
                elif record.args:
                    uu_sub = self.uuid_re.sub
                    args = tuple(
                        uu_sub(repl, str(a)) if isinstance(a, (str, uuid.UUID)) else a
                        for a in record.args
                    )
                    record.args = args
            except Exception as exc:
                # Basically, inform, but ignore
                print(f"Unable to colorize log message: {exc}")

        if record.levelno > logging.WARNING:
            return self._error_formatter.format(record)
        elif record.levelno > logging.INFO:
            return self._warning_formatter.format(record)
        elif record.levelno > logging.DEBUG:
            return self._info_formatter.format(record)
        return self._debug_formatter.format(record)


def _get_file_dict_config(
    logfile: str, console_enabled: bool, debug: bool, no_color: bool
) -> dict:
    # ensure that the logdir exists
    logdir = os.path.dirname(logfile)
    os.makedirs(logdir, exist_ok=True)
    log_handlers = ["logfile"]
    if console_enabled:
        log_handlers.append("console")

    return {
        "version": 1,
        "formatters": {
            "streamfmt": {
                "()": ComputeConsoleFormatter,
                "debug": debug,
                "no_color": no_color,
                "datefmt": "%Y-%m-%d %H:%M:%S,%f",
            },
            "filefmt": {
                "()": DatetimeFormatter,
                "format": DEFAULT_FORMAT,
                "datefmt": "%Y-%m-%d %H:%M:%S,%f",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "streamfmt",
            },
            "logfile": {
                "class": "logging.handlers.RotatingFileHandler",
                "level": "DEBUG",
                "filename": logfile,
                "formatter": "filefmt",
                "maxBytes": 100 * 1024 * 1024,
                "backupCount": 1,
            },
        },
        "loggers": {
            "globus_compute_endpoint": {
                "level": "DEBUG" if debug else "INFO",
                "handlers": log_handlers,
            },
            # configure for the Globus Compute SDK as well
            "globus_compute_sdk": {
                "level": "DEBUG" if debug else "WARNING",
                "handlers": log_handlers,
            },
        },
    }


def _get_stream_dict_config(debug: bool, no_color: bool) -> dict:
    return {
        "version": 1,
        "formatters": {
            "streamfmt": {
                "()": ComputeConsoleFormatter,
                "debug": debug,
                "no_color": no_color,
                "datefmt": "%Y-%m-%d %H:%M:%S,%f",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG" if debug else "INFO",
                "formatter": "streamfmt",
            }
        },
        "loggers": {
            "globus_compute_endpoint": {
                "level": "DEBUG",
                "handlers": ["console"],
            },
            # configure for the Globus Compute SDK as well
            "globus_compute_sdk": {
                "level": "DEBUG" if debug else "WARNING",
                "handlers": ["console"],
            },
        },
    }


class ComputeLogger(logging.Logger):
    TRACE = logging.DEBUG - 5

    def trace(self, msg, *args, **kwargs):
        self.log(ComputeLogger.TRACE, msg, args, **kwargs)


logging.setLoggerClass(ComputeLogger)
logger = logging.getLogger(__name__)


def setup_logging(
    *,
    logfile: pathlib.Path | str | None = None,
    console_enabled: bool = True,
    debug: bool = False,
    no_color: bool = False,
) -> None:
    if logfile is not None:
        config = _get_file_dict_config(str(logfile), console_enabled, debug, no_color)
    else:
        config = _get_stream_dict_config(debug, no_color)

    logging.config.dictConfig(config)



import logging
import os
import time
import typing as t
import uuid

from globus_compute_common import messagepack
from globus_compute_common.messagepack.message_types import Result, Task, TaskTransition
from globus_compute_common.tasks import ActorName, TaskState
from globus_compute_endpoint.engines.high_throughput.messages import Message
from globus_compute_endpoint.exception_handling import (
    get_error_string,
    get_result_error_details,
)
from globus_compute_endpoint.exceptions import CouldNotExecuteUserTaskError
from globus_compute_sdk.errors import MaxResultSizeExceeded
from globus_compute_sdk.sdk.utils import get_env_details
from globus_compute_sdk.serialize import ComputeSerializer
from parsl.app.python import timeout

log = logging.getLogger(__name__)

serializer = ComputeSerializer()


def execute_task(
    task_id: uuid.UUID, task_body: bytes, result_size_limit: int = 10 * 1024 * 1024
) -> bytes:
    """Execute task is designed to enable any executor to execute a Task payload
    and return a Result payload, where the payload follows the globus-compute protocols
    This method is placed here to make serialization easy for executor classes
    Parameters
    ----------
    task_id: uuid string
    task_body: packed message as bytes
    result_size_limit: result size in bytes
    Returns
    -------
    messagepack packed Result
    """
    exec_start = TaskTransition(
        timestamp=time.time_ns(), state=TaskState.EXEC_START, actor=ActorName.WORKER
    )

    result_message: dict[
        str,
        uuid.UUID | str | tuple[str, str] | list[TaskTransition] | dict[str, str],
    ]

    env_details = get_env_details()
    try:
        _task, task_buffer = _unpack_messagebody(task_body)
        log.debug("executing task task_id='%s'", task_id)
        result = _call_user_function(task_buffer, result_size_limit=result_size_limit)
        log.debug("Execution completed without exception")
        result_message = dict(task_id=task_id, data=result)

    except Exception:
        log.exception("Caught an exception while executing user function")
        code, user_message = get_result_error_details()
        error_details = {"code": code, "user_message": user_message}
        result_message = dict(
            task_id=task_id,
            data=get_error_string(),
            exception=get_error_string(),
            error_details=error_details,
        )

    result_message["details"] = env_details

    exec_end = TaskTransition(
        timestamp=time.time_ns(),
        state=TaskState.EXEC_END,
        actor=ActorName.WORKER,
    )

    result_message["task_statuses"] = [exec_start, exec_end]

    log.debug(
        "task %s completed in %d ns",
        task_id,
        (exec_end.timestamp - exec_start.timestamp),
    )

    return messagepack.pack(Result(**result_message))


def _unpack_messagebody(message: bytes) -> t.Tuple[Task, str]:
    """Unpack messagebody as a messagepack message with
    some legacy handling
    Parameters
    ----------
    message: messagepack'ed message body
    Returns
    -------
    tuple(task, task_buffer)
    """
    try:
        task = messagepack.unpack(message)
        if not isinstance(task, messagepack.message_types.Task):
            raise CouldNotExecuteUserTaskError(
                f"wrong type of message in worker: {type(task)}"
            )
        task_buffer = task.task_buffer
    # on parse errors, failover to trying the "legacy" message reading
    except (
        messagepack.InvalidMessageError,
        messagepack.UnrecognizedProtocolVersion,
    ):
        task = Message.unpack(message)
        assert isinstance(task, Task)
        task_buffer = task.task_buffer.decode("utf-8")  # type: ignore[attr-defined]
    return task, task_buffer


def _call_user_function(
    task_buffer: str, result_size_limit: int, serializer=serializer
) -> str:
    """Deserialize the buffer and execute the task.
    Parameters
    ----------
    task_buffer: serialized buffer of (fn, args, kwargs)
    result_size_limit: size limit in bytes for results
    serializer: serializer for the buffers
    Returns
    -------
    Returns serialized result or throws exception.
    """
    GC_TASK_TIMEOUT = max(0.0, float(os.environ.get("GC_TASK_TIMEOUT", 0.0)))
    f, args, kwargs = serializer.unpack_and_deserialize(task_buffer)
    if GC_TASK_TIMEOUT > 0.0:
        log.debug(f"Setting task timeout to GC_TASK_TIMEOUT={GC_TASK_TIMEOUT}s")
        f = timeout(f, GC_TASK_TIMEOUT)

    result_data = f(*args, **kwargs)
    serialized_data = serializer.serialize(result_data)

    if len(serialized_data) > result_size_limit:
        raise MaxResultSizeExceeded(len(serialized_data), result_size_limit)

    return serialized_data



class CouldNotExecuteUserTaskError(Exception):
    """
    generic exception class for errors while attempting to setup and run the user
    function but which does not come from the user function itself failing
    """



from __future__ import annotations

import logging
import pickle
import uuid

from globus_compute_common.messagepack import Message as OutgoingMessage
from globus_compute_common.messagepack import pack
from globus_compute_common.messagepack.message_types import (
    EPStatusReport as OutgoingEPStatusReport,
)
from globus_compute_common.messagepack.message_types import Result as OutgoingResult
from globus_compute_common.messagepack.message_types import (
    ResultErrorDetails as OutgoingResultErrorDetails,
)
from globus_compute_common.messagepack.message_types import Task as OutgoingTask
from globus_compute_common.messagepack.message_types import TaskTransition
from globus_compute_endpoint.engines.high_throughput.messages import (
    EPStatusReport as InternalEPStatusReport,
)
from globus_compute_endpoint.engines.high_throughput.messages import (
    Task as InternalTask,
)

logger = logging.getLogger(__name__)


def convert_ep_status_report(
    internal: InternalEPStatusReport,
) -> OutgoingEPStatusReport:
    messagepack_msg = OutgoingEPStatusReport(
        endpoint_id=internal._header,
        global_state=internal.global_state,
        task_statuses=internal.task_statuses,
    )
    return messagepack_msg


def try_convert_to_messagepack(message: bytes) -> bytes:
    try:
        unpacked = pickle.loads(message)
    except pickle.UnpicklingError:
        # message isn't pickled; assume that it's already in messagepack format
        return message

    messagepack_msg: OutgoingMessage | None = None

    if isinstance(unpacked, InternalEPStatusReport):
        messagepack_msg = OutgoingEPStatusReport(
            endpoint_id=unpacked._header,
            global_state=unpacked.global_state,
            task_statuses=unpacked.task_statuses,
        )
    elif isinstance(unpacked, dict):
        kwargs: dict[
            str, str | uuid.UUID | OutgoingResultErrorDetails | list[TaskTransition]
        ] = {
            "task_id": uuid.UUID(unpacked["task_id"]),
        }
        if "details" in unpacked:
            kwargs["details"] = unpacked["details"]
        if "task_statuses" in unpacked:
            kwargs["task_statuses"] = unpacked["task_statuses"]
        if "exception" in unpacked:
            kwargs["data"] = unpacked["exception"]
            code, user_message = unpacked.get("error_details", ("Unknown", "Unknown"))
            kwargs["error_details"] = OutgoingResultErrorDetails(
                code=code, user_message=user_message
            )
        else:
            kwargs["data"] = unpacked["data"]

        messagepack_msg = OutgoingResult(**kwargs)

    if messagepack_msg:
        message = pack(messagepack_msg)

    return message


def convert_to_internaltask(message: OutgoingTask, container_type: str | None) -> bytes:
    container_loc = "RAW"
    if message.container:
        for img in message.container.images:
            if img.image_type == container_type:
                container_loc = img.location
                break

    return InternalTask(
        task_id=str(message.task_id),
        container_id=container_loc,
        task_buffer=message.task_buffer,
    ).pack()



from __future__ import annotations

import multiprocessing
import unittest.mock
import uuid

from globus_compute_common import messagepack
from globus_compute_common.messagepack.message_types import Result, Task
from globus_compute_sdk import Client


class MockExecutor(unittest.mock.Mock):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.results_passthrough: multiprocessing.Queue | None = None
        self.passthrough = True

    def start(
        self,
        endpoint_id: uuid.UUID | None = None,
        run_dir: str | None = None,
        results_passthrough: multiprocessing.Queue = None,
        funcx_client: Client = None,
    ):
        self.results_passthrough = results_passthrough
        self.funcx_client = funcx_client
        self.endpoint_id = endpoint_id
        self.run_dir = run_dir

    def submit(self, task_id: uuid.UUID, packed_task: bytes):
        task: Task = messagepack.unpack(packed_task)
        res = Result(task_id=task_id, data=task.task_buffer)
        packed_result = messagepack.pack(res)
        msg = {"task_id": str(task_id), "message": packed_result}
        self.results_passthrough.put(msg)



from __future__ import annotations

import inspect
import warnings

from globus_compute_endpoint.engines import HighThroughputEngine
from parsl.utils import RepresentationMixin

_DEFAULT_EXECUTORS = [HighThroughputEngine()]


class Config(RepresentationMixin):
    """Specification of Globus Compute configuration options.

    Parameters
    ----------

    executors : list of Executors
        A list of executors which serve as the backend for function execution.
        As of 0.2.2, this list should contain only one executor.
        Default: [HighThroughputEngine()]

    environment: str
        Environment the endpoint should connect to. Sets funcx_service_address and
        results_ws_uri unless they are also specified. If not specified, the endpoint
        connects to production.
        Default: None

    funcx_service_address: str | None
        URL address string of the Globus Compute service to which the Endpoint
        should connect.
        Default: None

    results_ws_uri: str | None
        URL address string of the Globus Compute websocket service passed to the
        Globus Compute client.
        Default: None

    warn_about_url_mismatch: Bool
        Log a warning if funcx_service_address and results_ws_uri appear to point to
        different environments.
        Default: False

    heartbeat_period: int (seconds)
        The interval at which heartbeat messages are sent from the endpoint to the
        Globus Compute web service
        Default: 30s

    heartbeat_threshold: int (seconds)
        Seconds since the last hearbeat message from the Globus Compute web service
        after which the connection is assumed to be disconnected.
        Default: 120s

    idle_heartbeats_soft: int (count)
        Number of heartbeats after an endpoint is idle (no outstanding tasks or
        results, and at least 1 task or result has been forwarded) before the
        endpoint shuts down.  If 0, then the endpoint must be manually triggered
        to shut down (e.g., SIGINT or SIGTERM).
        Default: 0

    idle_heartbeats_hard: int (count)
        Number of heartbeats after no task or result has moved before the endpoint
        shuts down.  Unlike `idle_heartbeats_soft`, this idle timer does not require
        that there are no outstanding tasks or results.  If no task or result has
        moved in this many heartbeats, then the endpoint will shut down.  In
        particular, this is intended to catch the error condition that a worker
        has gone missing, and will thus _never_ return the task it was sent.
        Note that this setting is only enabled if the `idle_heartbeats_soft` is a
        value greater than 0.  Suggested value: a multiplier of heartbeat_period
        equivalent to two days.  For example, if `heartbeat_period` is 30s, then
        suggest 5760.
        Default: 5760

    stdout : str
        Path where the endpoint's stdout should be written
        Default: ./interchange.stdout

    stderr : str
        Path where the endpoint's stderr should be written
        Default: ./interchange.stderr

    detach_endpoint : Bool
        Should the endpoint deamon be run as a detached process? This is good for
        a real edge node, but an anti-pattern for kubernetes pods
        Default: True

    log_dir : str
        Optional path string to the top-level directory where logs should be written to.
        Default: None

    multi_tenant : bool | None
        Designates the endpoint as a multi-tenant endpoint
        Default: None

    allowed_functions : list[str] | None
        List of functions that are allowed to be run on the endpoint

    force_mt_allow_same_user : bool
        If set, override the heuristic that determines whether the uid running the
        multi-tenant endpoint may also run user endpoints.

        Normally, the multi-tenant endpoint disallows starting user endpoints with
        the same UID as the parent process unless the UID has no privileges.  That
        means that the UID is not 0 (root), or that the UID does *not* have (among
        many others) the capability to change the user (otherwise known as "drop
        privileges").

        Default: False

    display_name : str | None
        The display name for the endpoint.  If None, defaults to name
        Default: None

    endpoint_setup : str | None
        Command or commands to be run during the endpoint initialization process.

    endpoint_teardown : str | None
        Command or commands to be run during the endpoint shutdown process.
    """

    def __init__(
        self,
        # Execution backed
        executors: list = _DEFAULT_EXECUTORS,
        # Connection info
        environment: str | None = None,
        funcx_service_address=None,
        multi_tenant: bool | None = None,
        allowed_functions: list[str] | None = None,
        # Tuning info
        heartbeat_period=30,
        heartbeat_threshold=120,
        idle_heartbeats_soft=0,
        idle_heartbeats_hard=5760,  # Two days, divided by `heartbeat_period`
        detach_endpoint=True,
        endpoint_setup: str | None = None,
        endpoint_teardown: str | None = None,
        force_mt_allow_same_user: bool = False,
        # Misc info
        display_name: str | None = None,
        # Logging info
        log_dir=None,
        stdout="./endpoint.log",
        stderr="./endpoint.log",
        **kwargs,
    ):
        for deprecated_name in ("results_ws_uri", "warn_about_url_mismatch"):
            if deprecated_name in kwargs:
                caller_filename = inspect.stack()[1].filename
                msg = (
                    f"The '{deprecated_name}' argument is deprecated.  It will"
                    f" be removed in a future release.  Found in: {caller_filename}"
                )
                warnings.warn(msg)
                del kwargs[deprecated_name]

        for unknown_arg in kwargs:
            # Calculated multiple times, but only in errant case -- nominally
            # fixed and then "not an issue."
            caller_filename = inspect.stack()[1].filename
            cls_name = self.__class__.__name__
            msg = (
                f"Unknown argument to {cls_name} ignored: {unknown_arg}"
                f"\n  Specified in: {caller_filename}"
            )
            warnings.warn(msg)

        # Execution backends
        self.executors = executors  # List of executors

        # Connection info
        self.environment = environment
        self.funcx_service_address = funcx_service_address

        self.multi_tenant = multi_tenant is True
        self.force_mt_allow_same_user = force_mt_allow_same_user is True
        self.allowed_functions = allowed_functions

        # Tuning info
        self.heartbeat_period = heartbeat_period
        self.heartbeat_threshold = heartbeat_threshold
        self.idle_heartbeats_soft = int(max(0, idle_heartbeats_soft))
        self.idle_heartbeats_hard = int(max(0, idle_heartbeats_hard))
        self.detach_endpoint = detach_endpoint

        self.endpoint_setup = endpoint_setup
        self.endpoint_teardown = endpoint_teardown

        # Logging info
        self.log_dir = log_dir
        self.stdout = stdout
        self.stderr = stderr

        # Misc info
        self.display_name = display_name



from .globus_compute import GlobusComputeEngine
from .high_throughput.engine import HighThroughputEngine
from .process_pool import ProcessPoolEngine
from .thread_pool import ThreadPoolEngine

__all__ = (
    "GlobusComputeEngine",
    "ProcessPoolEngine",
    "ThreadPoolEngine",
    "HighThroughputEngine",
)



from globus_compute_endpoint.endpoint.config import Config
from globus_compute_endpoint.engines import HighThroughputEngine
from parsl.providers import LocalProvider

config = Config(
    display_name=None,  # If None, defaults to the endpoint name
    executors=[
        HighThroughputEngine(
            provider=LocalProvider(
                init_blocks=1,
                min_blocks=0,
                max_blocks=1,
            ),
        )
    ],
)

# For now, visible_to must be a list of URNs for globus auth users or groups, e.g.:
# urn:globus:auth:identity:{user_uuid}
# urn:globus:groups:id:{group_uuid}
meta = {
    "name": "$name",
    "description": "",
    "organization": "",
    "department": "",
    "public": False,
    "visible_to": [],
}



import os
from pathlib import Path

from setuptools import find_packages, setup

REQUIRES = [
    "requests>=2.31.0,<3",
    "globus-sdk",  # version will be bounded by `globus-compute-sdk`
    "globus-compute-sdk==2.4.0",
    "globus-compute-common==0.3.0a2",
    # table printing used in list-endpoints
    "texttable>=1.6.4,<2",
    # although psutil does not declare itself to use semver, it appears to offer
    # strong backwards-compatibility promises based on its changelog, usage, and
    # history
    #
    # TODO: re-evaluate bound after we have an answer of some kind from psutil
    # see:
    #   https://github.com/giampaolo/psutil/issues/2002
    "psutil<6",
    # provides easy daemonization of the endpoint
    "python-daemon>=2,<3",
    # CLI parsing
    "click>=8,<9",
    # disallow use of 22.3.0; the whl package on some platforms causes ZMQ issues
    #
    # NOTE: 22.3.0 introduced a patched version of libzmq.so to the wheel packaging
    # which may be the source of the problems , the problem can be fixed by
    # building from source, which may mean there's an issue in the packaged library
    # further investigation may be needed if the issue persists in the next pyzmq
    # release
    "pyzmq>=22.0.0,!=22.3.0,<=23.2.0",
    # 'parsl' is a core requirement of the globus-compute-endpoint, essential to a range
    # of different features and functions
    # pin exact versions because it does not use semver
    "parsl==2023.7.3",
    "pika>=1.2.0",
    "pyprctl<0.2.0",
    "setproctitle>=1.3.2,<1.4",
    "pyyaml>=6.0,<7.0",
    "jinja2>=3.1.2,<3.2",
    "jsonschema>=4.19.0,<4.20",
]

TEST_REQUIRES = [
    "responses",
    "pytest>=7.2",
    "coverage>=5.2",
    "pytest-mock==3.2.0",
    "pyfakefs",
]


version_ns = {}
with open(os.path.join("globus_compute_endpoint", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns["__version__"]

directory = Path(__file__).parent
long_description = (directory / "PyPI.md").read_text()

setup(
    name="globus-compute-endpoint",
    version=version,
    packages=find_packages(),
    description="Globus Compute: High Performance Function Serving for Science",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=REQUIRES,
    extras_require={
        "test": TEST_REQUIRES,
    },
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    keywords=["Globus Compute", "FaaS", "Function Serving"],
    entry_points={
        "console_scripts": [
            "globus-compute-endpoint=globus_compute_endpoint.cli:cli_run",
            "globus-compute-interchange"
            "=globus_compute_endpoint.engines.high_throughput.interchange:cli_run",
            "globus-compute-manager"
            "=globus_compute_endpoint.engines.high_throughput.manager:cli_run",
            "globus-compute-worker"
            "=globus_compute_endpoint.engines.high_throughput.worker:cli_run",
        ]
    },
    include_package_data=True,
    author="Globus Compute Team",
    author_email="support@globus.org",
    license="Apache License, Version 2.0",
    url="https://github.com/funcx-faas/funcx",
    project_urls={
        "Changelog": "https://globus-compute.readthedocs.io/en/latest/changelog.html",  # noqa: E501
        "Upgrade to Globus Compute": "https://globus-compute.readthedocs.io/en/latest/funcx_upgrade.html",  # noqa: E501
    },
)



from __future__ import annotations

import multiprocessing
import os
import random
import string

# multiprocessing.Event is a method, not a class
# to annotate, we need the "real" class
# see: https://github.com/python/typeshed/issues/4266
from multiprocessing.synchronize import Event as EventType

import pika
import pika.exceptions
import pytest
from globus_compute_endpoint.endpoint.rabbit_mq import (
    RabbitPublisherStatus,
    ResultPublisher,
    TaskQueueSubscriber,
)
from pika.exchange_type import ExchangeType
from tests.integration.test_rabbit_mq.result_queue_subscriber import (
    ResultQueueSubscriber,
)
from tests.integration.test_rabbit_mq.task_queue_publisher import TaskQueuePublisher
from tests.utils import try_assert


@pytest.fixture(scope="session")
def rabbitmq_conn_url():
    env_var_name = "RABBITMQ_INTEGRATION_TEST_URI"
    rmq_test_uri = os.getenv(env_var_name, "amqp:///")

    try:
        # Die here and now, first thing, with a hopefully-helpful direct fix suggestion
        # if rmq_test_uri is invalid or otherwise "not working."
        pika.BlockingConnection(pika.URLParameters(rmq_test_uri))
    except Exception as exc:
        msg = (
            f"Failed to connect to RabbitMQ via URI: {rmq_test_uri}\n"
            f"  Do you need to export {env_var_name} ?  Typo?"
        )
        raise ValueError(msg) from exc

    return rmq_test_uri


@pytest.fixture(scope="session")
def pika_conn_params(rabbitmq_conn_url):
    return pika.URLParameters(rabbitmq_conn_url)


def _flush_results(pika_conn_params):
    """Reminder: not a fixture; regular method"""
    with pika.BlockingConnection(pika_conn_params) as mq_conn:
        with mq_conn.channel() as chan:
            queue_name = "results"
            chan.exchange_declare(
                exchange="results",
                exchange_type="topic",
                durable=True,
            )
            chan.queue_declare(queue=queue_name, durable=True)
            chan.queue_purge(queue=queue_name)


@pytest.fixture
def flush_results(pika_conn_params):
    _flush_results(pika_conn_params)


@pytest.fixture(scope="session", autouse=True)
def clear_results_queue(pika_conn_params):
    _flush_results(pika_conn_params)
    yield
    _flush_results(pika_conn_params)


@pytest.fixture
def create_result_queue_info(rabbitmq_conn_url, tod_session_num, request):
    def _do_it(connection_url=None, queue_id=None) -> dict:
        exchange_name = "results"
        if not queue_id:
            queue_id = f"test_result_queue_{tod_session_num}__{request.node.name}"
        if not connection_url:
            connection_url = rabbitmq_conn_url
        routing_key = f"{queue_id}.results"
        return {
            "connection_url": connection_url,
            "exchange": exchange_name,
            "queue": "results",
            "queue_publish_kwargs": {
                "exchange": exchange_name,
                "routing_key": routing_key,
                "mandatory": True,
                "properties": {
                    "delivery_mode": pika.spec.PERSISTENT_DELIVERY_MODE,
                },
            },
            "test_routing_key": queue_id,
        }

    return _do_it


@pytest.fixture
def result_queue_info(create_result_queue_info) -> dict:
    return create_result_queue_info()


@pytest.fixture
def task_queue_info(rabbitmq_conn_url, tod_session_num, request) -> dict:
    queue_id = f"test_task_queue_{tod_session_num}__{request.node.name}"
    return {
        "connection_url": rabbitmq_conn_url,
        "exchange": "tasks",
        "queue": f"{queue_id}.tasks",
        "test_routing_key": queue_id,
    }


@pytest.fixture
def running_subscribers(request):
    run_list = []

    def cleanup():
        for x in run_list:
            try:  # cannot check is_alive on closed proc
                is_alive = x.is_alive()
            except ValueError:
                is_alive = False
            if is_alive:
                try:
                    x.stop()
                except Exception as e:
                    x.terminate()
                    raise Exception(
                        f"{x.__class__.__name__} did not shutdown correctly"
                    ) from e

    request.addfinalizer(cleanup)
    return run_list


@pytest.fixture(scope="session")
def ensure_result_queue(pika_conn_params):
    queues_created = []

    def _do_ensure(exchange_opts=None, queue_opts=None):
        if not exchange_opts:
            exchange_opts = {
                "exchange": "results",
                "exchange_type": ExchangeType.topic.value,
                "durable": True,
            }
        if not queue_opts:
            queue_opts = {"queue": "results", "durable": True}
            routing_key = "*.results"
        else:
            routing_key = queue_opts["queue"]

        with pika.BlockingConnection(pika_conn_params) as mq_conn:
            with mq_conn.channel() as chan:
                chan.exchange_declare(**exchange_opts)
                chan.queue_declare(**queue_opts)
                queues_created.append(queue_opts["queue"])
                chan.queue_bind(
                    queue=queue_opts["queue"],
                    exchange=exchange_opts["exchange"],
                    routing_key=routing_key,
                )

    _do_ensure()  # The main "results" should always exist for our tests
    yield _do_ensure

    with pika.BlockingConnection(pika_conn_params) as mq_conn:
        with mq_conn.channel() as chan:
            for q_name in queues_created:
                chan.queue_delete(q_name)


@pytest.fixture
def start_task_q_subscriber(
    running_subscribers,
    task_queue_info,
    default_endpoint_id,
    ensure_task_queue,
):
    def func(
        *,
        endpoint_id: str | None = None,
        queue: multiprocessing.Queue | None = None,
        quiesce_event: EventType | None = None,
        override_params: pika.connection.Parameters | None = None,
    ):
        if endpoint_id is None:
            endpoint_id = default_endpoint_id
        if quiesce_event is None:
            quiesce_event = multiprocessing.Event()
        if queue is None:
            queue = multiprocessing.Queue()
        q_info = task_queue_info if override_params is None else override_params
        ensure_task_queue(queue_opts={"queue": q_info["queue"]})

        task_q = TaskQueueSubscriber(
            queue_info=q_info,
            external_queue=queue,
            quiesce_event=quiesce_event,
            endpoint_id=endpoint_id,
        )
        task_q.start()
        running_subscribers.append(task_q)
        return task_q

    return func


@pytest.fixture
def start_result_q_subscriber(running_subscribers, pika_conn_params):
    def func(
        *,
        queue: multiprocessing.Queue | None = None,
        kill_event: multiprocessing.Event | None = None,
        override_params: pika.connection.Parameters | None = None,
    ):
        if kill_event is None:
            kill_event = multiprocessing.Event()
        if queue is None:
            queue = multiprocessing.Queue()
        result_q = ResultQueueSubscriber(
            conn_params=pika_conn_params if not override_params else override_params,
            external_queue=queue,
            kill_event=kill_event,
        )
        result_q.start()
        running_subscribers.append(result_q)
        if not result_q.test_class_ready.wait(10):
            raise AssertionError("Result Queue subscriber failed to initialize")
        return result_q

    return func


@pytest.fixture
def running_publishers(request):
    run_list = []

    def cleanup():
        for x in run_list:
            if x.status is RabbitPublisherStatus.connected:
                if hasattr(x, "stop"):
                    x.stop()  # ResultPublisher
                else:
                    x.close()  # TaskQueuePublisher (from tests)

    request.addfinalizer(cleanup)
    return run_list


@pytest.fixture
def start_result_q_publisher(
    running_publishers,
    result_queue_info,
    ensure_result_queue,
):
    def func(
        *,
        override_params: dict | None = None,
        queue_purge: bool = True,
    ):
        q_info = result_queue_info if override_params is None else override_params
        exchange_name, queue_name = q_info.get("exchange"), q_info.get("queue")
        if exchange_name:  # We'll call it an error to specify only one
            exchange_opts = {
                "exchange": exchange_name,
                "exchange_type": ExchangeType.topic.value,
                "durable": True,
            }
            queue_opts = {"queue": queue_name, "durable": True}
            ensure_result_queue(exchange_opts=exchange_opts, queue_opts=queue_opts)

        result_pub = ResultPublisher(queue_info=q_info)
        result_pub.start()
        if queue_purge:  # Make sure queue is empty
            try_assert(lambda: result_pub._mq_chan is not None)
            result_pub._mq_chan.queue_purge(q_info["queue"])
        running_publishers.append(result_pub)
        return result_pub

    return func


@pytest.fixture
def start_task_q_publisher(
    running_publishers,
    task_queue_info,
    ensure_task_queue,
    default_endpoint_id,
):
    def func(
        *,
        override_params: pika.connection.Parameters | None = None,
        queue_purge: bool = True,
    ):
        q_info = task_queue_info if override_params is None else override_params
        exchange_name, queue_name = q_info.get("exchange"), q_info.get("queue")
        if exchange_name:  # We'll call it an error to specify only one
            exchange_opts = {
                "exchange": exchange_name,
                "exchange_type": ExchangeType.direct.value,
            }
            queue_opts = {"queue": queue_name, "arguments": {"x-expires": 30 * 1000}}
            ensure_task_queue(exchange_opts=exchange_opts, queue_opts=queue_opts)

        task_pub = TaskQueuePublisher(queue_info=q_info)
        task_pub.connect()
        if queue_purge:  # Make sure queue is empty
            task_pub._channel.queue_purge(q_info["queue"])
        running_publishers.append(task_pub)
        return task_pub

    return func


@pytest.fixture(scope="session")
def ensure_task_queue(pika_conn_params, tod_session_num, request):
    queues_created = []

    def _do_ensure(exchange_opts=None, queue_opts=None):
        if not exchange_opts:
            exchange_opts = {
                "exchange": "tasks",
                "exchange_type": ExchangeType.direct.value,
            }
        exchange_opts.setdefault("durable", True)
        if not queue_opts:
            rndm = "".join(random.choice(string.ascii_letters) for _ in range(5))
            queue_id = f"queue_{tod_session_num}__{request.node.name}__{rndm}"
            queue_opts = {"queue": f"task_{queue_id}.tasks"}
        # play nice with dev/test resources; auto clean
        queue_opts.setdefault("arguments", {"x-expires": 30 * 1000})

        with pika.BlockingConnection(pika_conn_params) as mq_conn:
            with mq_conn.channel() as chan:
                chan.exchange_declare(**exchange_opts)
                chan.queue_declare(**queue_opts)
                queues_created.append(queue_opts["queue"])
                chan.queue_bind(
                    queue=queue_opts["queue"],
                    exchange=exchange_opts["exchange"],
                    routing_key=queue_opts["queue"],
                )

    yield _do_ensure

    with pika.BlockingConnection(pika_conn_params) as mq_conn:
        with mq_conn.channel() as chan:
            for q_name in queues_created:
                chan.queue_delete(q_name)



"""
common elements to rabbitmq tools
"""

import enum


class RabbitPublisherStatus(enum.Enum):
    closed = enum.auto()
    connected = enum.auto()


# a status enum to describe the different states of the subscriber subprocs
class SubscriberProcessStatus(enum.Enum):
    # the object is in the parent process, it has no notion of the process state other
    # than that it is none of the other values
    parent = enum.auto()
    # the process is in the child proc and should try to start or continue to run
    running = enum.auto()
    # the process is in the child proc and is closing down, it should not try to
    # reconnect or otherwise handle failures
    closing = enum.auto()



import logging
import os
import uuid

import zmq
import zmq.auth
from zmq.auth.thread import ThreadAuthenticator

log = logging.getLogger(__name__)


class TaskQueue:
    """Outgoing task queue from the executor to the Interchange"""

    def __init__(
        self,
        address: str,
        port: int = 55001,
        identity: str = str(uuid.uuid4()),
        zmq_context=None,
        set_hwm=False,
        RCVTIMEO=None,
        SNDTIMEO=None,
        linger=None,
        ironhouse: bool = False,
        keys_dir: str = os.path.abspath(".curve"),
        mode: str = "client",
    ):
        """
        Parameters
        ----------

        address: str
           address to connect

        port: int
           Port to use

        identity : str
           Applies only to clients, where the identity must match the endpoint uuid.
           This will be utf-8 encoded on the wire. A random uuid4 string is set by
           default.

        mode: string
           Either 'client' or 'server'

        keys_dir : string
           Directory from which keys will be loaded for curve.

        ironhouse: Bool
           Only valid for server mode. Setting this flag switches the server to require
           client keys to be available on the server in the keys_dir.
        """
        if zmq_context:
            self.context = zmq_context
        else:
            self.context = zmq.Context()

        self.mode = mode
        self.port = port
        self.ironhouse = ironhouse
        self.keys_dir = keys_dir

        assert self.mode in [
            "client",
            "server",
        ], "Only two modes are supported: client, server"

        if self.mode == "server":
            print("Configuring server")
            self.zmq_socket = self.context.socket(zmq.ROUTER)
            self.zmq_socket.set(zmq.ROUTER_MANDATORY, 1)
            self.zmq_socket.set(zmq.ROUTER_HANDOVER, 1)
            print("Setting up auth-server")
            self.setup_server_auth()
        elif self.mode == "client":
            self.zmq_socket = self.context.socket(zmq.DEALER)
            self.setup_client_auth()
            self.zmq_socket.setsockopt(zmq.IDENTITY, identity.encode("utf-8"))
        else:
            raise ValueError(
                "TaskQueue must be initialized with mode set to 'server' or 'client'"
            )

        if set_hwm:
            self.zmq_socket.set_hwm(0)
        if RCVTIMEO is not None:
            self.zmq_socket.setsockopt(zmq.RCVTIMEO, RCVTIMEO)
        if SNDTIMEO is not None:
            self.zmq_socket.setsockopt(zmq.SNDTIMEO, SNDTIMEO)
        if linger is not None:
            self.zmq_socket.setsockopt(zmq.LINGER, linger)

        # all zmq setsockopt calls must be done before bind/connect is called
        if self.mode == "server":
            self.zmq_socket.bind(f"tcp://*:{port}")
        elif self.mode == "client":
            self.zmq_socket.connect(f"tcp://{address}:{port}")

        self.poller = zmq.Poller()
        self.poller.register(self.zmq_socket)
        os.makedirs(self.keys_dir, exist_ok=True)
        log.debug(f"Initializing Taskqueue:{self.mode} on port:{self.port}")

    def zmq_context(self):
        return self.context

    def add_client_key(self, endpoint_id, client_key):
        log.info("Adding client key")
        if self.ironhouse:
            # Use the ironhouse ZMQ pattern: http://hintjens.com/blog:49#toc6
            with open(os.path.join(self.keys_dir, f"{endpoint_id}.key"), "w") as f:
                f.write(client_key)
            try:
                self.auth.configure_curve(domain="*", location=self.keys_dir)
            except Exception:
                log.exception("Failed to load keys from {self.keys_dir}")
        return

    def setup_server_auth(self):
        # Start an authenticator for this context.
        self.auth = ThreadAuthenticator(self.context)
        self.auth.start()
        self.auth.allow("127.0.0.1")
        # Tell the authenticator how to handle CURVE requests

        if not self.ironhouse:
            # Use the stonehouse ZMQ pattern: http://hintjens.com/blog:49#toc5
            self.auth.configure_curve(domain="*", location=zmq.auth.CURVE_ALLOW_ANY)

        server_secret_file = os.path.join(self.keys_dir, "server.key_secret")
        server_public, server_secret = zmq.auth.load_certificate(server_secret_file)
        self.zmq_socket.curve_secretkey = server_secret
        self.zmq_socket.curve_publickey = server_public
        self.zmq_socket.curve_server = True  # must come before bind

    def setup_client_auth(self):
        # We need two certificates, one for the client and one for
        # the server. The client must know the server's public key
        # to make a CURVE connection.
        client_secret_file = os.path.join(self.keys_dir, "endpoint.key_secret")
        client_public, client_secret = zmq.auth.load_certificate(client_secret_file)
        self.zmq_socket.curve_secretkey = client_secret
        self.zmq_socket.curve_publickey = client_public

        # The client must know the server's public key to make a CURVE connection.
        server_public_file = os.path.join(self.keys_dir, "server.key")
        server_public, _ = zmq.auth.load_certificate(server_public_file)
        self.zmq_socket.curve_serverkey = server_public

    def get(self, block=True, timeout=1000):
        """
        Parameters
        ----------

        block : Bool
            Blocks until there's a message, Default is True
        timeout : int
            Milliseconds to wait.
        """
        # timeout is in milliseconds
        if block is True:
            return self.zmq_socket.recv_multipart()

        socks = dict(self.poller.poll(timeout=timeout))
        if self.zmq_socket in socks and socks[self.zmq_socket] == zmq.POLLIN:
            message = self.zmq_socket.recv_multipart()
            return message
        else:
            raise zmq.Again

    def register_client(self, message):
        return self.zmq_socket.send_multipart([message])

    def put(self, dest, message, max_timeout=1000):
        """This function needs to be fast at the same time aware of the possibility of
        ZMQ pipes overflowing.

        The timeout increases slowly if contention is detected on ZMQ pipes.
        We could set copy=False and get slightly better latency but this results
        in ZMQ sockets reaching a broken state once there are ~10k tasks in flight.
        This issue can be magnified if each the serialized buffer itself is larger.

        Parameters
        ----------

        dest : zmq_identity of the destination endpoint, must be a byte string

        message : py object
             Python object to send

        max_timeout : int
             Max timeout in milliseconds that we will wait for before raising an
             exception

        Raises
        ------

        zmq.EAGAIN if the send failed.
        zmq.error.ZMQError: Host unreachable (if client disconnects?)

        """
        if self.mode == "client":
            return self.zmq_socket.send_multipart([message])
        else:
            return self.zmq_socket.send_multipart([dest, message])

    def close(self):
        self.zmq_socket.close()
        self.context.term()



import queue
import time

from globus_compute_endpoint.strategies import SimpleStrategy


class MockInterchange:
    def __init__(self, max_blocks=1, tasks=10):
        self.tasks_pending = tasks
        self.max_blocks = max_blocks
        self.managers = 0
        self.status = self.create_data()

    def get_outstanding_breakdown(self):
        item = self.status.get()

        if self.tasks_pending >= self.managers:
            this_round = self.tasks_pending - self.managers
            self.tasks_pending -= self.managers
        else:
            this_round = self.tasks_pending
            self.tasks_pending = 0

        current = [("interchange", this_round, this_round)]
        for i in range(self.managers):
            current.extend((f"manager_{i}", 1, 1))
        self.status.put(current)

        return item

    def scale_out(self):
        self.managers += 1

    def create_data(self):
        q = queue.Queue()
        items = [
            [("interchange", 0, 0)],
            [("interchange", 0, 0)],
            [("interchange", 0, 0)],
            [("interchange", self.tasks_pending, self.tasks_pending)],
            [("interchange", self.tasks_pending, self.tasks_pending)],
        ]
        [q.put(i) for i in items]

        return q


if __name__ == "__main__":
    print("Starting")
    mock = MockInterchange()
    # strategy = BaseStrategy(mock, threshold=2, interval=1)
    strategy = SimpleStrategy(mock, threshold=2, interval=1)
    print(strategy)
    for i in range(10):
        if i < 3:
            mock.scale_out()
        time.sleep(1)
    print("Exiting now")



# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/
__version__ = "2.4.0"

# TODO: remove after a `globus-compute-sdk` release
# this is needed because it's imported by `globus-compute-sdk` to do the version check
VERSION = __version__

# Here as it's the easier way for funcx-endpoint cli to display it
DEPRECATION_FUNCX_ENDPOINT = """
funcX Endpoint has been renamed to Globus Compute Endpoint and the new package
is available on PyPI:
    https://pypi.org/project/globus-compute-endpoint/

Please consider upgrading to Globus Compute.  More information can be found at:
    https://globus-compute.readthedocs.io/en/latest/funcx_upgrade.html
"""


# app name to send as part of requests
app_name = f"Globus Compute Endpoint v{__version__}"



from __future__ import annotations

import argparse
import logging
import os
import signal
import sys
import typing as t

import dill
import zmq
from globus_compute_common import messagepack
from globus_compute_common.messagepack.message_types import Result as OutgoingResult
from globus_compute_common.messagepack.message_types import (
    ResultErrorDetails as OutgoingResultErrorDetails,
)
from globus_compute_endpoint.engines.high_throughput.messages import Message
from globus_compute_endpoint.exception_handling import (
    get_error_string,
    get_result_error_details,
)
from globus_compute_endpoint.logging_config import setup_logging
from globus_compute_sdk.sdk.utils import get_env_details
from globus_compute_sdk.serialize import ComputeSerializer

log = logging.getLogger(__name__)

DEFAULT_RESULT_SIZE_LIMIT_MB = 10
DEFAULT_RESULT_SIZE_LIMIT_B = DEFAULT_RESULT_SIZE_LIMIT_MB * 1024 * 1024


class Worker:
    """The Globus Compute worker
    Parameters
    ----------

    worker_id : str
     Worker id string

    address : str
     Address at which the manager might be reached. This is usually 127.0.0.1

    port : int
     Port at which the manager can be reached

    result_size_limit : int
     Maximum result size allowed in Bytes
     Default = 10 MB

    Globus Compute worker will use the REP sockets to:
         task = recv ()
         result = execute(task)
         send(result)
    """

    def __init__(
        self,
        worker_id,
        address,
        port,
        worker_type="RAW",
        result_size_limit=DEFAULT_RESULT_SIZE_LIMIT_B,
    ):
        self.worker_id = worker_id
        self.address = address
        self.port = port
        self.worker_type = worker_type
        self.serializer = ComputeSerializer()
        self.serialize = self.serializer.serialize
        self.deserialize = self.serializer.deserialize
        self.result_size_limit = result_size_limit

        log.info(f"Initializing worker {worker_id}")
        log.info(f"Worker is of type: {worker_type}")

        self.context = zmq.Context()
        self.poller = zmq.Poller()
        self.identity = worker_id.encode()

        self.task_socket = self.context.socket(zmq.DEALER)
        self.task_socket.setsockopt(zmq.IDENTITY, self.identity)

        log.info(f"Trying to connect to : tcp://{self.address}:{self.port}")
        self.task_socket.connect(f"tcp://{self.address}:{self.port}")
        self.poller.register(self.task_socket, zmq.POLLIN)
        signal.signal(signal.SIGTERM, self.handler)

    def handler(self, signum, frame):
        log.error(f"Signal handler called with signal {signum}")
        sys.exit(1)

    def _send_registration_message(self):
        log.debug("Sending registration")
        payload = {"worker_id": self.worker_id, "worker_type": self.worker_type}
        self.task_socket.send_multipart([b"REGISTER", dill.dumps(payload)])

    def start(self):
        log.info("Starting worker")
        self._send_registration_message()

        while True:
            log.debug("Waiting for task")
            p_task_id, p_container_id, msg = self.task_socket.recv_multipart()
            task_id: str = dill.loads(p_task_id)
            container_id: str = dill.loads(p_container_id)
            log.debug(f"Received task with task_id='{task_id}' and msg='{msg}'")

            if task_id == "KILL":
                log.info("[KILL] -- Worker KILL message received! ")
                # send a "worker die" message back to the manager
                self.task_socket.send_multipart([b"WRKR_DIE", b""])
                log.info(f"*** WORKER {self.worker_id} ABOUT TO DIE ***")
                # Kill the worker after accepting death in message to manager.
                sys.exit()
            else:
                result = self._worker_execute_task(task_id, msg)
                result["container_id"] = container_id
                log.debug("Sending result")
                # send bytes over the socket back to the manager
                self.task_socket.send_multipart([b"TASK_RET", dill.dumps(result)])

        log.warning("Broke out of the loop... dying")

    def compose_exception_message(self, task_id: str) -> bytes:
        code, user_message = get_result_error_details()
        outgoing_result = OutgoingResult(
            task_id=task_id,
            data=get_error_string(),
            error_details=OutgoingResultErrorDetails(
                code=code,
                user_message=user_message,
            ),
            details=get_env_details(),
        )
        return messagepack.pack(outgoing_result)

    def _worker_execute_task(
        self, task_id: str, msg: bytes
    ) -> dict[str, t.Union[str, bytes]]:
        result_message: dict[str, t.Union[str, bytes]] = {"task_id": task_id}
        try:
            # Unwrap HTEX's Task packing
            task_message = Message.unpack(msg)
            serialized_fn_package = task_message.task_buffer.decode()

            # Deserialize HTEX Engines' wrapping of
            # execute_task, messagepack_payload)
            function, args, kwargs = self.deserialize(serialized_fn_package)

            # Execute
            serialized_result: bytes = function(*args, **kwargs)
            result_message["data"] = serialized_result

        except Exception:
            log.exception("Failed to execute task")
            serialized_error = self.compose_exception_message(task_id)
            result_message["data"] = serialized_error

        return result_message


def cli_run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w", "--worker_id", required=True, help="ID of worker from process_worker_pool"
    )
    parser.add_argument(
        "-t", "--type", required=False, help="Container type of worker", default="RAW"
    )
    parser.add_argument(
        "-a", "--address", required=True, help="Address for the manager, eg X,Y,"
    )
    parser.add_argument(
        "-p",
        "--port",
        required=True,
        help="Internal port at which the worker connects to the manager",
    )
    parser.add_argument(
        "--logdir", required=True, help="Directory path where worker log files written"
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Directory path where worker log files written",
    )
    args = parser.parse_args()

    setup_logging(
        logfile=os.path.join(args.logdir, f"funcx_worker_{args.worker_id}.log"),
        debug=args.debug,
    )

    # Redirect the stdout and stderr
    stdout_path = os.path.join(args.logdir, f"funcx_worker_{args.worker_id}.stdout")
    stderr_path = os.path.join(args.logdir, f"funcx_worker_{args.worker_id}.stderr")
    with open(stdout_path, "w") as fo, open(stderr_path, "w") as fe:
        # Redirect the stdout
        old_stdout, old_stderr = sys.stdout, sys.stderr
        sys.stdout = fo
        sys.stderr = fe

        try:
            worker = Worker(
                args.worker_id,
                args.address,
                int(args.port),
                worker_type=args.type,
            )
            worker.start()
        finally:
            # Switch them back
            sys.stdout = old_stdout
            sys.stderr = old_stderr


if __name__ == "__main__":
    cli_run()
